#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.readers.bed`

    :py:class:`BED_Reader`
        Reads BED files to SegmentChain objects

See http://genome.ucsc.edu/FAQ/FAQformat.html
"""
import functools
import warnings
import pandas as pd
import os

from csv import QUOTE_NONE
from nose.plugins.attrib import attr
from plastid.util.services.mini2to3 import cStringIO
from plastid.genomics.roitools import SegmentChain, GenomicSegment, Transcript
from plastid.readers.bed import BED_Reader
from nose.tools import assert_equal, assert_true, assert_dict_equal, assert_greater_equal

from plastid.test.ref_files import RPATH, REF_FILES

warnings.simplefilter("ignore", DeprecationWarning)

TXBED = os.path.join(RPATH, REF_FILES["100transcripts_bed"])
CDSBED = os.path.join(RPATH, REF_FILES["100cds_bed"])

from plastid.test.data.annotations.py100cds import control_cds
from plastid.test.data.annotations.py100transcripts import control_transcripts


#===============================================================================
# INDEX: test suites
#===============================================================================

@attr(test="unit")
class TestBED():
    """Test case for BED input/output"""

    @classmethod
    def setUpClass(cls):
        cls.header = _BED_HEADER
        cls.data = {}
        cls.extracol_data = {}
        bed_df = pd.read_table(
            cStringIO.StringIO(_BED12_DATA), header=None, sep="\t", index_col=None
        )
        extra_df = pd.read_table(
            cStringIO.StringIO(_EXTRA_COLS), header=0, sep="\t", index_col=None
        )
        cls.big_df = pd.concat([bed_df, extra_df], axis=1)

        for n in (3, 4, 5, 6, 8, 9, 12):
            cls.data[n] = cls.get_bed_subset(cls.header, n, 0)
            cls.extracol_data[n] = cls.get_bed_subset(cls.header, n, 4)

    @classmethod
    def get_bed_subset(cls, header, bed_cols, extra_cols=0):
        buf = cStringIO.StringIO()
        columns = cls.big_df.columns[list(range(bed_cols)) + list(range(12, 12 + extra_cols))]
        cls.big_df.to_csv(
            buf, columns=columns, sep="\t", index=False, header=False, quoting=QUOTE_NONE
        )  #,float_format="%.8f")
        return buf.getvalue()

    @staticmethod
    def check_equal(found, expected, msg=None):
        if msg is not None:
            assert_equal(found, expected, msg)
        else:
            assert_equal(found, expected)

    def open_from_str(self):
        for expected, found in zip(control_transcripts, BED_Reader(TXBED, return_type=Transcript)):
            assert_equal(expected, found)

    def open_from_multi_str(self):
        for expected, found in zip(control_transcripts + control_cds,
                                   BED_Reader(TXBED, CDSBED, return_type=Transcript)):
            assert_equal(expected, found)

    def open_from_fh(self):
        with open(TXBED) as fh:
            for expected, found in zip(control_transcripts, BED_Reader(fh, return_type=Transcript)):
                assert_equal(expected, found)

    def open_from_multi_fh(self):
        with open(TXBED) as fh:
            with open(CDSBED) as fh2:
                for expected, found in zip(control_transcripts + control_cds,
                                           BED_Reader(fh, fh2, return_type=Transcript)):
                    assert_equal(expected, found)

    def test_bed_import_3to12plus4_columns_with_formatters(self):
        names = [
            ("numcol", int),
            ("floatcol", float),
            ("strcol", str),
            ("attrcol", str),
        ]

        tx_reader = functools.partial(BED_Reader, return_type=Transcript, extra_columns=names)
        seg_reader = functools.partial(BED_Reader, return_type=SegmentChain, extra_columns=names)
        tests = [
            (seg_reader, _TEST_SEGMENTCHAINS, "reader_segmentchain"),
            (tx_reader, _TEST_TRANSCRIPTS, "reader_transcript"),
        ]
        for reader_fn, known_set, name in tests:
            for n, data_str in sorted(self.extracol_data.items()):
                c = 0
                for (test_ivc, known_ivc) in zip(reader_fn(cStringIO.StringIO(data_str)),
                                                 known_set):
                    for x in range(4):
                        colname = names[x][0]
                        assert_true(
                            colname in test_ivc.attr,
                            "Column name '%s' not found in attr dict (%s BED columns)" % (x, n)
                        )
                        assert_equal(test_ivc.attr[colname], self.big_df.iloc[c, 12 + x])

                    # columns: chrom, start, end
                    if n >= 3:
                        # no strand info, so we need to test iv.start, iv.end, iv.chrom
                        err_msg = "%s failed endpoint equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        yield self.check_equal, known_ivc.spanning_segment.start, test_ivc.spanning_segment.start, err_msg
                        yield self.check_equal, known_ivc.spanning_segment.end, test_ivc.spanning_segment.end, err_msg
                        yield self.check_equal, known_ivc.spanning_segment.chrom, test_ivc.spanning_segment.chrom, err_msg
                    # column: name
                    if n >= 4:
                        err_msg = "%s failed name equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc.attr, test_ivc.attr
                        )
                        yield self.check_equal, known_ivc.attr["ID"], test_ivc.attr["ID"], err_msg
                    # column: score
                    if n >= 5:
                        err_msg = "%s failed score equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc.attr, test_ivc.attr
                        )
                        yield self.check_equal, known_ivc.attr.get(
                            "score", 0
                        ), test_ivc.attr["score"], err_msg
                    # column : strand
                    if n >= 6:
                        err_msg = "%s failed strand equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        yield self.check_equal, known_ivc.spanning_segment.strand, test_ivc.spanning_segment.strand
                    # column: color
                    if n >= 9:
                        err_msg = "%s failed color equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc.attr, test_ivc.attr
                        )
                        yield self.check_equal, known_ivc.attr.get(
                            "color", "#000000"
                        ), test_ivc.attr["color"], err_msg
                    # columns: exon/block info
                    if n == 12:
                        err_msg = "%s failed block equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        for iv1, iv2 in zip(known_ivc, test_ivc):
                            assert_equal(iv1, iv2, err_msg)
                        err_msg = "%s failed position set on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        yield self.check_equal, known_ivc.get_position_set(
                        ), test_ivc.get_position_set(), err_msg

                    c += 1

                yield self.check_equal, c, len(
                    known_set
                ), "Not all intervals loaded! Expected %s, found %s." % (len(known_set), c)

    def test_bed_import_3to12plus4_columns_with_names(self):
        names = [X for X in self.big_df.columns[-4:]]
        tx_reader = functools.partial(BED_Reader, return_type=Transcript, extra_columns=names)
        seg_reader = functools.partial(BED_Reader, return_type=SegmentChain, extra_columns=names)
        tests = [
            (seg_reader, _TEST_SEGMENTCHAINS, "reader_segmentchain"),
            (tx_reader, _TEST_TRANSCRIPTS, "reader_transcript"),
        ]
        for reader_fn, known_set, name in tests:
            for n, data_str in sorted(self.extracol_data.items()):
                c = 0
                for (test_ivc, known_ivc) in zip(reader_fn(cStringIO.StringIO(data_str)),
                                                 known_set):
                    for x in range(4):
                        colname = names[x]
                        assert_true(colname in test_ivc.attr)
                        assert_equal(str(test_ivc.attr[colname]), str(self.big_df.iloc[c, 12 + x]))

                    # columns: chrom, start, end
                    if n >= 3:
                        # no strand info, so we need to test iv.start, iv.end, iv.chrom
                        err_msg = "%s failed endpoint equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        yield self.check_equal, known_ivc.spanning_segment.start, test_ivc.spanning_segment.start, err_msg
                        yield self.check_equal, known_ivc.spanning_segment.end, test_ivc.spanning_segment.end, err_msg
                        yield self.check_equal, known_ivc.spanning_segment.chrom, test_ivc.spanning_segment.chrom, err_msg
                    # column: name
                    if n >= 4:
                        err_msg = "%s failed name equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc.attr, test_ivc.attr
                        )
                        yield self.check_equal, known_ivc.attr["ID"], test_ivc.attr["ID"], err_msg
                    # column: score
                    if n >= 5:
                        err_msg = "%s failed score equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc.attr, test_ivc.attr
                        )
                        yield self.check_equal, known_ivc.attr.get(
                            "score", 0
                        ), test_ivc.attr["score"], err_msg
                    # column : strand
                    if n >= 6:
                        err_msg = "%s failed strand equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        yield self.check_equal, known_ivc.spanning_segment.strand, test_ivc.spanning_segment.strand
                    # column: color
                    if n >= 9:
                        err_msg = "%s failed color equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc.attr, test_ivc.attr
                        )
                        yield self.check_equal, known_ivc.attr.get(
                            "color", "#000000"
                        ), test_ivc.attr["color"], err_msg
                    # columns: exon/block info
                    if n == 12:
                        err_msg = "%s failed block equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        for iv1, iv2 in zip(known_ivc, test_ivc):
                            assert_equal(iv1, iv2, err_msg)
                        err_msg = "%s failed position set on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        yield self.check_equal, known_ivc.get_position_set(
                        ), test_ivc.get_position_set(), err_msg

                    c += 1

                yield self.check_equal, c, len(
                    known_set
                ), "Not all intervals loaded! Expected %s, found %s." % (len(known_set), c)

    def test_bed_import_3to12plus4_columns_with_int(self):
        tx_reader = functools.partial(BED_Reader, return_type=Transcript, extra_columns=4)
        seg_reader = functools.partial(BED_Reader, return_type=SegmentChain, extra_columns=4)
        tests = [
            (seg_reader, _TEST_SEGMENTCHAINS, "reader_segmentchain"),
            (tx_reader, _TEST_TRANSCRIPTS, "reader_transcript"),
        ]
        for reader_fn, known_set, name in tests:
            for n, data_str in sorted(self.extracol_data.items()):
                c = 0
                for (test_ivc, known_ivc) in zip(reader_fn(cStringIO.StringIO(data_str)),
                                                 known_set):
                    for x in range(4):
                        colname = "custom%s" % x
                        assert_true(colname in test_ivc.attr)
                        assert_equal(str(test_ivc.attr[colname]), str(self.big_df.iloc[c, 12 + x]))

                    # columns: chrom, start, end
                    if n >= 3:
                        # no strand info, so we need to test iv.start, iv.end, iv.chrom
                        err_msg = "%s failed endpoint equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        yield self.check_equal, known_ivc.spanning_segment.start, test_ivc.spanning_segment.start, err_msg
                        yield self.check_equal, known_ivc.spanning_segment.end, test_ivc.spanning_segment.end, err_msg
                        yield self.check_equal, known_ivc.spanning_segment.chrom, test_ivc.spanning_segment.chrom, err_msg
                    # column: name
                    if n >= 4:
                        err_msg = "%s failed name equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc.attr, test_ivc.attr
                        )
                        yield self.check_equal, known_ivc.attr["ID"], test_ivc.attr["ID"], err_msg
                    # column: score
                    if n >= 5:
                        err_msg = "%s failed score equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc.attr, test_ivc.attr
                        )
                        yield self.check_equal, known_ivc.attr.get(
                            "score", 0
                        ), test_ivc.attr["score"], err_msg
                    # column : strand
                    if n >= 6:
                        err_msg = "%s failed strand equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        yield self.check_equal, known_ivc.spanning_segment.strand, test_ivc.spanning_segment.strand
                    # column: color
                    if n >= 9:
                        err_msg = "%s failed color equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc.attr, test_ivc.attr
                        )
                        yield self.check_equal, known_ivc.attr.get(
                            "color", "#000000"
                        ), test_ivc.attr["color"], err_msg
                    # columns: exon/block info
                    if n == 12:
                        err_msg = "%s failed block equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        for iv1, iv2 in zip(known_ivc, test_ivc):
                            assert_equal(iv1, iv2, err_msg)
                        err_msg = "%s failed position set on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        yield self.check_equal, known_ivc.get_position_set(
                        ), test_ivc.get_position_set(), err_msg

                    c += 1

                yield self.check_equal, c, len(
                    known_set
                ), "Not all intervals loaded! Expected %s, found %s." % (len(known_set), c)

    def test_bed_export_3to12plus4_columns_with_names(self):
        names = [X for X in self.big_df.columns[-4:]]
        tests = [
            (Transcript.from_bed, _TEST_TRANSCRIPTS, "tx_frombed_plus4_int"),
            (SegmentChain.from_bed, _TEST_SEGMENTCHAINS, "segchain_frombed_plus4_int"),
        ]
        for import_fn, known_set, name in tests:
            extracol12plus = [X.split("\t") for X in self.extracol_data[12].strip("\n").split("\n")]
            for n, data_str in sorted(self.extracol_data.items()):
                for c, line in enumerate(data_str.strip("\n").split("\n")):
                    out_line = import_fn(line, extra_columns=names).as_bed(as_int=False).strip("\n")
                    out_items = out_line.split("\t")[:n] + out_line.split("\t")[-4:]
                    expected_items = extracol12plus[c][:n] + extracol12plus[c][-4:]
                    msg = "%s BED %s+%s export unequal for lines:\nin:  %s\nout: %s\nexp: %s" % (
                        name, n, 4, line, "\t".join(out_items), "\t".join(expected_items)
                    )
                    yield self.check_equal, out_items, expected_items, msg

                yield self.check_equal, c + 1, len(
                    known_set
                ), "Not all intervals loaded! Expected %s, found %s." % (len(known_set), c)

    def test_bed_export_3to12plus4_columns_with_int(self):
        tests = [
            (Transcript.from_bed, _TEST_TRANSCRIPTS, "tx_frombed_plus4_int"),
            (SegmentChain.from_bed, _TEST_SEGMENTCHAINS, "segchain_frombed_plus4_int"),
        ]
        for import_fn, known_set, name in tests:
            extracol12plus = [X.split("\t") for X in self.extracol_data[12].strip("\n").split("\n")]
            for n, data_str in sorted(self.extracol_data.items()):
                for c, line in enumerate(data_str.strip("\n").split("\n")):
                    out_line = import_fn(line, extra_columns=4).as_bed(as_int=False).strip("\n")
                    out_items = out_line.split("\t")[:n] + out_line.split("\t")[-4:]
                    expected_items = extracol12plus[c][:n] + extracol12plus[c][-4:]
                    msg = "%s BED %s+%s export unequal for lines:\nin:  %s\nout: %s\nexp: %s" % (
                        name, n, 4, line, "\t".join(out_items), "\t".join(expected_items)
                    )
                    yield self.check_equal, out_items, expected_items, msg

                yield self.check_equal, c + 1, len(
                    known_set
                ), "Not all intervals loaded! Expected %s, found %s." % (len(known_set), c)

    def test_bed_import_3to12_columns(self):
        tx_reader = functools.partial(BED_Reader, return_type=Transcript)
        tests = [
            (BED_Reader, _TEST_SEGMENTCHAINS, "reader_segmentchain"),
            (tx_reader, _TEST_TRANSCRIPTS, "reader_transcript"),
        ]
        for reader_fn, known_set, name in tests:
            for n, data_str in sorted(self.data.items()):
                c = 0
                for (test_ivc, known_ivc) in zip(reader_fn(cStringIO.StringIO(data_str)),
                                                 known_set):
                    # columns: chrom, start, end
                    if n >= 3:
                        # no strand info, so we need to test iv.start, iv.end, iv.chrom
                        err_msg = "%s failed endpoint equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        yield self.check_equal, known_ivc.spanning_segment.start, test_ivc.spanning_segment.start, err_msg
                        yield self.check_equal, known_ivc.spanning_segment.end, test_ivc.spanning_segment.end, err_msg
                        yield self.check_equal, known_ivc.spanning_segment.chrom, test_ivc.spanning_segment.chrom, err_msg
                    # column: name
                    if n >= 4:
                        err_msg = "%s failed name equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc.attr, test_ivc.attr
                        )
                        yield self.check_equal, known_ivc.attr["ID"], test_ivc.attr["ID"], err_msg
                    # column: score
                    if n >= 5:
                        err_msg = "%s failed score equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc.attr, test_ivc.attr
                        )
                        yield self.check_equal, known_ivc.attr.get(
                            "score", 0
                        ), test_ivc.attr["score"], err_msg
                    # column : strand
                    if n >= 6:
                        err_msg = "%s failed strand equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        yield self.check_equal, known_ivc.spanning_segment.strand, test_ivc.spanning_segment.strand
                    # column: color
                    if n >= 9:
                        err_msg = "%s failed color equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc.attr, test_ivc.attr
                        )
                        yield self.check_equal, known_ivc.attr.get(
                            "color", "#000000"
                        ), test_ivc.attr["color"], err_msg
                    # columns: exon/block info
                    if n == 12:
                        err_msg = "%s failed block equality on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        for iv1, iv2 in zip(known_ivc, test_ivc):
                            assert_equal(iv1, iv2, err_msg)
                        err_msg = "%s failed position set on %s-column BED input: %s,%s" % (
                            name, n, known_ivc, test_ivc
                        )
                        yield self.check_equal, known_ivc.get_position_set(
                        ), test_ivc.get_position_set(), err_msg

                    c += 1

                yield self.check_equal, c, len(
                    known_set
                ), "Not all intervals loaded! Expected %s, found %s." % (len(known_set), c)

    def test_ivcollection_thick_start_end_8to12_columns(self):
        """Checks equality of thickstart and thickend attributes for SegmentChain objects"""
        for n, data_str in sorted(self.data.items()):
            for c, (test_ivc, known_ivc) in enumerate(zip(BED_Reader(
                    cStringIO.StringIO(data_str), return_type=SegmentChain), _TEST_SEGMENTCHAINS)):
                if n >= 8:
                    err_msg = "Failed thickstart/end equality on %s-column BED input: %s,%s" % (
                        n, known_ivc.attr, test_ivc.attr
                    )
                    if known_ivc.attr.get("thickstart", None) is not None:
                        yield self.check_equal, known_ivc.attr["thickstart"], test_ivc.attr[
                            "thickstart"
                        ], err_msg
                    if known_ivc.attr.get("thickend", None) is not None:
                        yield self.check_equal, known_ivc.attr.get("thickend"), test_ivc.attr[
                            "thickend"
                        ], err_msg

            yield self.check_equal, c, 20 - 1, "Not all intervals loaded! Expected %s, found %s." % (
                20 - 1, c
            )

    def test_transcript_cds_start_end_8to12_columns(self):
        """Checks equality of endpoints of coding regions for Transcript objects"""
        for n, data_str in sorted(self.data.items()):
            for c, (test_ivc, known_ivc) in enumerate(zip(BED_Reader(
                    cStringIO.StringIO(data_str), return_type=Transcript), _TEST_TRANSCRIPTS)):
                if n >= 8:
                    err_msg = "Failed thickstart/end equality on %s-column BED input: %s,%s" % (
                        n, known_ivc.attr, test_ivc.attr
                    )
                    if known_ivc.attr.get("cds_genome_start", None) is not None:
                        yield self.check_equal, known_ivc.attr["cds_start"], test_ivc.attr[
                            "cds_start"
                        ], err_msg
                        yield self.check_equal, known_ivc.attr["cds_genome_start"], test_ivc.attr[
                            "cds_genome_start"
                        ], err_msg
                        yield self.check_equal, known_ivc.cds_genome_start, test_ivc.cds_genome_start, err_msg
                        yield self.check_equal, known_ivc.cds_start, test_ivc.cds_start, err_msg
                    if known_ivc.attr.get("cds_genome_end", None) is not None:
                        yield self.check_equal, known_ivc.attr["cds_end"], test_ivc.attr["cds_end"
                                                                                         ], err_msg
                        yield self.check_equal, known_ivc.attr["cds_genome_end"], test_ivc.attr[
                            "cds_genome_end"
                        ], err_msg
                        yield self.check_equal, known_ivc.cds_genome_end, test_ivc.cds_genome_end, err_msg
                        yield self.check_equal, known_ivc.cds_end, test_ivc.cds_end, err_msg

            yield self.check_equal, c, 20 - 1, "Not all intervals loaded! Expected %s, found %s." % (
                20 - 1, c
            )

    def test_track_subtype_parsing(self):
        reader = BED_Reader(cStringIO.StringIO(_NARROW_PEAK_TEXT))
        for c, (found, expected) in enumerate(zip(reader, _NARROW_PEAK_CHAINS)):
            found.attr.pop("color")
            found.attr.pop("score")
            assert_dict_equal(found.attr, expected.attr)
            assert_equal(found, expected)

        assert_equal(c, len(_NARROW_PEAK_CHAINS) - 1)

    def test_track_subtype_raises_warning_if_wrong_extra_columns(self):
        reader = BED_Reader(cStringIO.StringIO(_NARROW_PEAK_TEXT), extra_columns=14)
        with warnings.catch_warnings(record=True) as warns:
            warnings.simplefilter("always")
            ltmp = list(reader)
            assert_greater_equal(len(warns), 0)


#===============================================================================
# INDEX: test data
#===============================================================================

# test dataset, constructed manually to include various edge cases
_TEST_SEGMENTCHAINS = [
    # single-interval
    SegmentChain(GenomicSegment("chrA", 100, 1100, "+"), ID="IVC1p"),
    SegmentChain(GenomicSegment("chrA", 100, 1100, "-"), ID="IVC1m"),
    # multi-interval
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "+"), GenomicSegment("chrA", 2100, 2600, "+"), ID="IVC2p"
    ),
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "-"), GenomicSegment("chrA", 2100, 2600, "-"), ID="IVC2m"
    ),
    # multi-interval, with score
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "+"),
        GenomicSegment("chrA", 2100, 2600, "+"),
        ID="IVC3p",
        score=500
    ),
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "-"),
        GenomicSegment("chrA", 2100, 2600, "-"),
        ID="IVC3m",
        score=500
    ),
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "+"),
        GenomicSegment("chrA", 2100, 2600, "+"),
        GenomicSegment("chrA", 2605, 2700, "+"),
        ID="IVC4p",
        score=500
    ),
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "-"),
        GenomicSegment("chrA", 2100, 2600, "-"),
        ID="IVC4m",
        score=500
    ),
    # multi-interval, with score and color
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "+"),
        GenomicSegment("chrA", 2100, 2600, "+"),
        GenomicSegment("chrA", 2605, 2700, "+"),
        ID="IVC5p",
        score=500,
        color="#007ADF"
    ),
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "-"),
        GenomicSegment("chrA", 2100, 2600, "-"),
        GenomicSegment("chrA", 2605, 2700, "-"),
        ID="IVC5m",
        score=500,
        color="#007ADF"
    ),
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "+"),
        GenomicSegment("chrA", 2100, 2600, "+"),
        GenomicSegment("chrA", 2605, 2700, "+"),
        ID="IVC6p",
        score=500,
        color="#007ADF"
    ),
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "-"),
        GenomicSegment("chrA", 2100, 2600, "-"),
        GenomicSegment("chrA", 2605, 2700, "-"),
        ID="IVC6m",
        score=500,
        color="#007ADF"
    ),
    # multi-interval, with score, color, thickstart, and thickend, internally
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "+"),
        GenomicSegment("chrA", 2100, 2600, "+"),
        GenomicSegment("chrA", 2605, 2700, "+"),
        ID="IVC7p",
        score=500,
        color="#007ADF",
        thickstart=2200,
        thickend=2400
    ),
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "-"),
        GenomicSegment("chrA", 2100, 2600, "-"),
        GenomicSegment("chrA", 2605, 2700, "-"),
        ID="IVC7m",
        score=500,
        color="#007ADF",
        thickstart=2200,
        thickend=2400
    ),
    # multi-interval, thickend and thickstart covering whole SegmentChain
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "+"),
        GenomicSegment("chrA", 2100, 2600, "+"),
        GenomicSegment("chrA", 2605, 2700, "+"),
        ID="IVC8p",
        score=500,
        color="#007ADF",
        thickstart=100,
        thickend=2700
    ),
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "-"),
        GenomicSegment("chrA", 2100, 2600, "-"),
        GenomicSegment("chrA", 2605, 2700, "-"),
        ID="IVC8m",
        score=500,
        color="#007ADF",
        thickstart=100,
        thickend=2700
    ),
    # multi-interval, thickend and thickstart at exon-exon junctions
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "+"),
        GenomicSegment("chrA", 2100, 2600, "+"),
        GenomicSegment("chrA", 2605, 2700, "+"),
        ID="IVC9p",
        score=500,
        color="#007ADF",
        thickstart=2100,
        thickend=2600
    ),
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "-"),
        GenomicSegment("chrA", 2100, 2600, "-"),
        GenomicSegment("chrA", 2605, 2700, "-"),
        ID="IVC9m",
        score=500,
        color="#007ADF",
        thickstart=2100,
        thickend=2600
    ),
    # multi-interval, thickend and thickstart at exon-exon junctions
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "+"),
        GenomicSegment("chrA", 2100, 2600, "+"),
        GenomicSegment("chrA", 2605, 2700, "+"),
        ID="IVC10p",
        score=500,
        color="#007ADF",
        thickstart=1099,
        thickend=2101
    ),
    SegmentChain(
        GenomicSegment("chrA", 100, 1100, "-"),
        GenomicSegment("chrA", 2100, 2600, "-"),
        GenomicSegment("chrA", 2605, 2700, "-"),
        ID="IVC10m",
        score=500,
        color="#007ADF",
        thickstart=1099,
        thickend=2101
    ),
]

# same data, as transcripts
_TEST_TRANSCRIPTS = [Transcript(*X.segments, **X.attr) for X in _TEST_SEGMENTCHAINS]

_BED_HEADER = """browser position chrA:100-1100
track name=test_data description='my test data'
"""

# same data, as BED12 block
_BED12_DATA = """chrA    100    1100    IVC1p    0.0    +    100    100    0,0,0    1    1000,    0,
chrA    100    1100    IVC1m    0.0    -    100    100    0,0,0    1    1000,    0,
chrA    100    2600    IVC2p    0.0    +    100    100    0,0,0    2    1000,500,    0,2000,
chrA    100    2600    IVC2m    0.0    -    100    100    0,0,0    2    1000,500,    0,2000,
chrA    100    2600    IVC3p    500.0    +    100    100    0,0,0    2    1000,500,    0,2000,
chrA    100    2600    IVC3m    500.0    -    100    100    0,0,0    2    1000,500,    0,2000,
chrA    100    2700    IVC4p    500.0    +    100    100    0,0,0    3    1000,500,95,    0,2000,2505,
chrA    100    2600    IVC4m    500.0    -    100    100    0,0,0    2    1000,500,    0,2000,
chrA    100    2700    IVC5p    500.0    +    100    100    0,122,223    3    1000,500,95,    0,2000,2505,
chrA    100    2700    IVC5m    500.0    -    100    100    0,122,223    3    1000,500,95,    0,2000,2505,
chrA    100    2700    IVC6p    500.0    +    100    100    0,122,223    3    1000,500,95,    0,2000,2505,
chrA    100    2700    IVC6m    500.0    -    100    100    0,122,223    3    1000,500,95,    0,2000,2505,
chrA    100    2700    IVC7p    500.0    +    2200    2400    0,122,223    3    1000,500,95,    0,2000,2505,
chrA    100    2700    IVC7m    500.0    -    2200    2400    0,122,223    3    1000,500,95,    0,2000,2505,
chrA    100    2700    IVC8p    500.0    +    100    2700    0,122,223    3    1000,500,95,    0,2000,2505,
chrA    100    2700    IVC8m    500.0    -    100    2700    0,122,223    3    1000,500,95,    0,2000,2505,
chrA    100    2700    IVC9p    500.0    +    2100    2600    0,122,223    3    1000,500,95,    0,2000,2505,
chrA    100    2700    IVC9m    500.0    -    2100    2600    0,122,223    3    1000,500,95,    0,2000,2505,
chrA    100    2700    IVC10p    500.0    +    1099    2101    0,122,223    3    1000,500,95,    0,2000,2505,
chrA    100    2700    IVC10m    500.0    -    1099    2101    0,122,223    3    1000,500,95,    0,2000,2505,""".replace(
    "    ", "\t"
)

_EXTRA_COLS = """numcol    floatcol    strcol    attrcol
0    3.14    a    gene_id "gene_0"; transcript_id "transcript_0";
1    2.72523    abc    gene_id "gene_1"; transcript_id "transcript_1";
2    30.12350    DEF    gene_id "gene_2"; transcript_id "transcript_2";
3    15123.20    ghi    gene_id "gene_3"; transcript_id "transcript_3";
4    2.0    alongword    gene_id "gene_4"; transcript_id "transcript_4";
5    -3.1234    a sentence with spaces    gene_id "gene_5"; transcript_id "transcript_5";
6    -20.5    some notes with "quotes"    gene_id "gene_6"; transcript_id "transcript_6";
7    -1e10    1    gene_id "gene_7"; transcript_id "transcript_7";
8    2e5    2    gene_id "gene_8"; transcript_id "transcript_8";
9    2.3e6    3.0    gene_id "gene_9"; transcript_id "transcript_9";
10    0.03    string1    gene_id "gene_10"; transcript_id "transcript_10";
11    1.0    string2    gene_id "gene_11"; transcript_id "transcript_11";
12    2.0    string3    gene_id "gene_12"; transcript_id "transcript_12";
13    3.0    string4 string5 string6    gene_id "gene_13"; transcript_id "transcript_13";
14    4.0    test    gene_id "gene_14"; transcript_id "transcript_14";
15    5.0    testetst    gene_id "gene_15"; transcript_id "transcript_15";
16    6.0    testsatsdfasf    gene_id "gene_16"; transcript_id "transcript_16";
17    7.0    asdgahghfzgdasdfasdf    gene_id "gene_17"; transcript_id "transcript_17";
18    8.0    asdfasdfadsfgaasdg    gene_id "gene_18"; transcript_id "transcript_18";
19    9.0    asdfasdfdasfdas    gene_id "gene_19"; transcript_id "transcript_19";
""".replace("    ", "\t")

_NARROW_PEAK_TEXT = """track type=narrowPeak
chrI    100    15000    feature1    0    +    341.2    -123.2    -513.3    50
chrII    320    15000    feature2    0    -    2.1    -5123.2    0    650""".replace("    ", "\t")

_NARROW_PEAK_CHAINS = [
    SegmentChain(
        GenomicSegment("chrI", 100, 15000, "+"),
        ID='feature1',
        signalValue=341.2,
        pValue=-123.2,
        qValue=-513.3,
        peak=50,
        _bedx_column_order=["signalValue", "pValue", "qValue", "peak"],
        thickstart=100,
        thickend=100
    ),
    SegmentChain(
        GenomicSegment("chrII", 320, 15000, "-"),
        ID='feature2',
        signalValue=2.1,
        pValue=-5123.2,
        qValue=0.0,
        peak=650,
        _bedx_column_order=["signalValue", "pValue", "qValue", "peak"],
        thickstart=320,
        thickend=320
    ),
]



#!/usr/bin/env python
"""Utility classes, analagous to Unix-style pipes, for filtering or processing
input or output streams, such as file objects.

These filters may be composed by wrapping one around another, to perform
multiple operations in line on, for example, a text stream, before finally
converting it to another type of data.

Various readers are included:

    :class:`AbstractReader`
        Base class for all Readers. To create a Reader, subclass this and
        override the :py:meth:`~AbstractReader.filter` method.
    
    :class:`CommentReader`
        read through text documents, skipping over commented lines
    
    :class:`SkipBlankReader`
        read through text documents, skipping over blank lines
    
    :class:`FunctionReader`
        apply an arbitrary function to each unit of input data
    
    :class:`TeeReader`
        similar to Unix/Linux shell command ``tee``. Allows input streams to be
        sent to an arbitrary number of listeners

And various writers:

    :class:`AbstractWriter`
        Base class for all writers. To create a Writer, subclass this and
        override the :py:meth:`~AbstractReader.filter` method.

    :class:`ColorWriter`
        Enable ANSI coloring of text to output streams that support color.
        For streams that do not support color, text is not colored.
    
    :class:`NameDateWriter`
        Prepend timestamps to each line of string input before writing

    :class:`CommentWriter`
        Filter out commented lines from text stream before writing


And one convenience function:

    :func:`colored`
        Colorize text (via :func:`termcolor.colored`) if and only
        if color is supported by :obj:`sys.stderr` 
        

Examples
--------
Open a file, skipping comments and blank lines::

    >>> my_reader = CommentReader(SkipBlankReader(open("some_file.txt")))
    >>> for line in my_reader:
    >>>     pass # do something with each line, now that comments are removed

Open a file, applying a function ``foo_func`` to each line of input (*note,*
``foo_func`` can return any data type, not just a string)::

    >>> my_reader = FunctionReader(open("some_file.txt"),foo_func)
    >>> for data_unit in my_reader:
    >>>     pass # do something with each `foo`ed data unit

Write to stdout, prepending name and date::

    >>> import sys
    >>> my_writer = NameDateWriter(stream=sys.stdout)
    >>> my_writer.write(some_text)

"""
from __future__ import print_function
import sys
import datetime
from abc import abstractmethod
from io import IOBase

import termcolor

# color detection hint from http://stackoverflow.com/questions/7445658/how-to-detect-if-the-console-does-support-ansi-escape-codes-in-python
if hasattr(sys.stderr, "isatty") and sys.stderr.isatty():
    colored = termcolor.colored
else:
    colored = lambda x, **kwargs: str(x)

#===============================================================================
# INDEX: readers
#===============================================================================


class AbstractReader(IOBase):
    """Abstract base class for stream-reading filters. These may be wrapped around
    open-file like objects, for example, to remove comments or blank lines from text,
    or to convert units of input from one data type to another.
    
    Create a filter by subclassing this, and defining `self.filter()`
    
    See also
    --------
    CommentReader
        A reader that removes comments from text data
    
    """

    def __init__(self, stream):
        """Create an |AbstractReader|
        
        Parameters
        ----------
        stream : file-like
            Input data
        """
        self.stream = stream

    def isatty(self):
        return hasattr(self.stream, "isatty") and self.stream.isatty()

    def writable(self):
        return False

    def seekable(self):
        return False

    def readable(self):
        return True

    def fileno(self):
        raise IOError()

    def __next__(self):
        return self.filter(next(self.stream))

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def read(self):
        """Similar to :py:func:`file.read`. Process all units of data, assuming it is string-like
        
        Returns
        -------
        str
        """
        return "".join(self.readlines())

    def readline(self):
        """Process a single line of data, assuming it is string-like
        ``next(self)`` is more likely to behave as expected.
        
        Returns
        -------
        object
            a unit of processed data
        """
        return self.filter(self.stream.readline())

    def readlines(self):
        """Similar to :py:func:`file.readlines`.
        
        Returns
        -------
        list
            processed data
        """
        lreturn = []
        for line in self:
            lreturn.append(line)

        return lreturn

    def close(self):
        """Close stream"""
        try:
            self.stream.close()
        except AttributeError:
            pass

    @abstractmethod
    def filter(self, data):
        """Method that filters or processes each unit of data.
        Override this in subclasses
        
        Parameters
        ----------
        data : unit of data
            Whatever data to filter/format. Often string, but not necessary
        
        Returns
        -------
        object
            formatted data. Often string, but not necessarily
        """
        pass


class FunctionReader(AbstractReader):
    """Apply a function to each line in an input stream
    
        
    Parameters
    ----------
    stream : file-like
        Input stream
        
    func : function
        Function to apply to each unit of input in `stream`    
    """

    def __init__(self, stream, func):
        """Create a FunctionReader
        
        Parameters
        ----------
        stream : file-like
            Input stream
            
        func : function
            Function to apply to each unit of input in `stream`
        """
        self.filter = func
        AbstractReader.__init__(self, stream)


class SkipBlankReader(AbstractReader):
    """Ignores blank/whitespace-only lines in a text stream"""

    def filter(self, line):
        """Return next non-blank line of text
        
        Parameters
        ----------
        line : str
            Line of text

        Returns
        -------
        str
        """
        if len(line.strip()) == 0:
            return self.__next__()
        else:
            return line


class CommentReader(AbstractReader):
    """Ignore lines beginning with `'#'`, optionally preceded by whitespace
    from a text stream. Comments beginning mid line are left in-place
    """

    def __init__(self, stream):
        """Create a CommentReader
        
        Parameters
        ----------
        stream : file-like
            Input data
        """
        self.comments = []
        AbstractReader.__init__(self, stream)

    def get_comments(self):
        """Return all of the comments that have been found so far.
        Does not remove comments from the internal list.

        Returns
        -------
        list
            Comments found in text
        """
        return self.comments

    def filter(self, line):
        """Return next non-commented line of text
        
        Parameters
        ----------
        line : str
            Line of text

        Returns
        -------
        str
        """
        ltmp = line.lstrip()
        if len(ltmp) > 1 and ltmp[0] == "#":
            self.comments.append(line.strip())
            return self.__next__()
        else:
            return line


class BackwardReader(AbstractReader):
    """Reverses each line of a text stream character-wise."""

    def filter(self, line):
        """Return next non-commented line of text
        
        Parameters
        ----------
        line : str
            Line of text

        Returns
        -------
        str
            Reversed line of text
        """
        return line[::-1]


class TeeReader(AbstractReader):
    """Tee an input stream to listeners that register themselves with the TeeReader
    via the add_listener() method. Similar to shell command ``tee``

    Each listener must implement an ``alert()`` method, in order to receive the data.
    If `alert()` is not implemented, errors are suppressed.
    
    See also
    --------
    TeeListener : an example of a listener class
    """

    def __init__(self, stream):
        """Create an TeeReader
        
        Parameters
        ----------
        stream : file-like
            Input data
        """
        self.listeners = []
        AbstractReader.__init__(self, stream)

    def add_listener(self, listener):
        """Register a single listener with this reader
        
        Parameters
        ----------
        listener : |TeeListener|-like
        """
        self.listeners.append(listener)

    def add_listeners(self, *many_listeners):
        """Register one or more listeners with this reader
        
        Parameters
        ----------
        many_listeners : one or more |TeeListener|-like
        """
        for listener in many_listeners:
            self.add_listener(listener)

    def filter(self, line):
        """Sends each line to each listener. Complains if listener cannot listen!
        
        Parameters
        ----------
        line : a unit of input data
        
        Returns
        -------
        input data
        """
        for listener in self.listeners:
            try:
                listener.alert(line)
            except AttributeError:
                import warnings
                warnings.warn("Could not alert listener %s: " % str(listener))
        return line


class TeeListener(object):
    """Listener class for TeeFilter. Listeners if registered with a |TeeReader|
    will receive and process each unit of input via its ``alert()`` method"""

    @abstractmethod
    def alert(self, data):
        """Process input from a |TeeReader|.
        Override this method to perform the appropriate behavior.
        
        Parameters
        ----------
        data : object
            A unit of data
        """
        pass


class TestTeeListener(TeeListener):
    """Example of a TeeListener"""

    def alert(self, line):
        print(self.name + " heard something: " + line)


#===============================================================================
# INDEX: writers
#===============================================================================


class AbstractWriter(IOBase):
    """Abstract base class for stream-writing filters.
    Create a filter by subclassing this, and defining self.filter().
    
    Inherits `isatty()` from `self.stream`
        
    Parameters
    ----------
    stream : file-like, open for writing
        Output stream to which filtered/formatted data will be written    
    """

    def __init__(self, stream):
        """Create an AbstractWriter
        
        Parameters
        ----------
        stream : file-like, open for writing
            Output stream to which filtered/formatted data will be written
        """
        self.stream = stream

    def isatty(self):
        return hasattr(self.stream, "isatty") and self.stream.isatty()

    def writable(self):
        return True

    def seekable(self):
        return False

    def readable(self):
        return False

    def fileno(self):
        raise IOError()

    def write(self, data):
        """Write data to `self.stream`
        
        Parameters
        ----------
        data : unit of data
            Whatever data to filter/format. Often string, but not necessary
        """
        self.stream.write(self.filter(data))

    def flush(self):
        """Flush `self.stream`"""
        self.stream.flush()

    def close(self):
        """flush and close `self.stream`"""
        try:
            self.flush()
            self.stream.close()
        except:
            pass

    @abstractmethod
    def filter(self, data):
        """Method that filters or processes each unit of data.
        Override this in subclasses
        
        Parameters
        ----------
        data : unit of data
            Whatever data to filter/format. Often string, but not necessarily
        
        Returns
        -------
        object
            formatted data. Often string, but not necessary
        """
        pass


class ColorWriter(AbstractWriter):
    """Detect whether output stream supports color, and enable/disable colored output
        
    Parameters
    ----------
    stream : file-like
        Stream to write to (Default: :obj:`sys.stderr`)
    
    Methods
    -------
    :meth:`color`
        Color text. Delegates to :func:`termcolor.colored` if color is supported.
        Otherwise, returns uncolored text.
    """

    def __init__(self, stream=None):
        """Create a ColorWriter
        
        Parameters
        ----------
        stream : file-like
            Stream to write to (Default: :obj:`sys.stderr`)
        """
        AbstractWriter.__init__(self, stream=stream)
        if hasattr(self.stream, "isatty") and self.stream.isatty():
            self.color = termcolor.colored

    def color(self, text, **kwargs):
        """Color `text` with attributes specified in `kwargs` if `stream` supports ANSI color.
        
        See :func:`termcolor.colored` for usage
        
        Returns
        -------
        str
            `text`, colored as indicated, if color is supported
        """
        return text


class NameDateWriter(ColorWriter):
    """Prepend program name, date, and time to each line of output"""

    def __init__(self, name, line_delimiter="\n", stream=None):
        """Create a NameDateWriter
        
        Parameters
        ----------
        name : str
            Name to prepend
        
        stream : file-like
            Stream to write to (Default: :obj:`sys.stderr`)
        
        line_delimiter : str, optional
            Delimiter, postpended to lines. (Default `'\n'`)
        """
        stream = sys.stderr if stream is None else stream
        ColorWriter.__init__(self, stream=stream)
        self.name = name
        self.delimiter = line_delimiter
        self.fmtstr = "%s %s%s %s%s: {2}%s" % (
            self.color(name, color="blue", attrs=["bold"]),
            self.color("[", color="blue", attrs=["bold"]), self.color("{0}", color="green"),
            self.color("{1}", color="green", attrs=["bold"]),
            self.color("]", color="blue", attrs=["bold"]), self.delimiter
        )

    def filter(self, line):
        """Prepend date and time to each line of input
        
        Parameters
        ----------
        line : str
            Input
        
        Returns
        -------
        str : Input with date and time prepended
        """
        now = datetime.datetime.now()
        d = datetime.datetime.strftime(now, "%Y-%m-%d")
        t = datetime.datetime.strftime(now, "%T")
        return self.fmtstr.format(d, t, line.strip(self.delimiter))

    # included for backward compatibility
    def __call__(self, line):
        self.write(line)


class CommentWriter(AbstractWriter):
    """Filter out lines beginning with `'#'` from data written to a stream"""

    def filter(self, line):
        if line.startswith("#"):
            return ""
        else:
            return line


# alias included for backward compatibility
Printer = NameDateWriter



cdef enum Strand:
    error_strand   = 666
    undef_strand   = 0
    forward_strand = 1   # can use bitwise operations to detect overlap
    reverse_strand = 2
    unstranded     = 3

cdef enum ExBool:
    bool_exception = -1
    false     = 0
    true      = 1

cdef str strand_to_str(Strand)
cdef Strand str_to_strand(str) except error_strand

cdef class _GeneratorWrapper(object):
    cdef:
        object generator
        str    desc



# -*- coding: utf-8 -*-
#
# plastid documentation build configuration file, created by
# sphinx-quickstart on Fri Dec  5 11:55:54 2014.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys
import unittest
import datetime
import mock
import plastid

# -- General configuration ------------------------------------------------

project = u'plastid'
copyright = u'2014-2022, Joshua G. Dunn'
version = "0.6.1"


# set up substitutions file for automated crossreferences
rst_prolog = """
.. include:: /class_substitutions.txt
.. include:: /script_substitutions.txt
.. include:: /links.txt
"""

master_doc = 'index'

# ignore package prefix when alphabetizing index
modindex_common_prefix = ["plastid."]

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'numpydoc',
    'sphinxcontrib.bibtex',
#    'sphinxcontrib.argdoc',
]

html_static_path = ["_static"]
bibtex_bibfiles = ["zreferences.bib"]

# theming, compatibility both for local and builds on readthedocs -------------

# code from http://read-the-docs.readthedocs.org/en/latest/theme.html
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# sphinx autodoc config -------------------------------------------------------

autodoc_mock_imports = ["pandas", "scipy", "twobitreader"]

autodoc_default_options = {
    "undoc-members" : None,
    "special-members" : None,
    "inherited-members" : None,
}

autodoc_member_order = "groupwise"

autodoc_docstring_signature = True

numpydoc_class_members_toctree = False

# never document these methods/attributes
exclude_always = {
    "__class__",
    "__delattr__",
    "__dict__",
    "__format__",
    "__hash__",
    "__iter__",
    "__module__",
    "__new__",
    "__reduce__",
    "__reduce_ex__",
    "__repr__",
    "__sizeof__",
    "__subclasshook__",
    "__weakref__",
    "__del__"
} # yapf: disable

# document these only if their docstrings don't match those of their base classes
exclude_if_no_redoc_base_classes = [object, unittest.TestCase]
exclude_if_no_redoc = {
    "__init__",
    "__getattribute__",
    "__setattribute__",
    "__getattr__",
    "__setattr__",
    "__getitem__",
    "__setitem__",
    "__str__",
    "next",
    "__next__",
    "close",
}
exclude_if_no_redoc |= set(dir(unittest.TestCase))


def autodoc_skip_member(app, what, name, obj, skip, options):
    """Do not generate documentation for functions/methods/classes/objects that
    either:
    
        1.  Appear above ``exclude_always``

        2.  Have an attribute called ``plastid_skipdoc`` set to *True*
        
        3.  Have the same docstring as the method with the same name defined in 
            a base class that appears in ``exclude_if_no_redoc_base_classes``
    
    Parameters
    ----------
    app
        Sphinx application
    
    what : str
        Type of object (e.g. "module", "function", "class")
    
    name : str
        Name of object within module scope
    
    obj : object
        Object to skip or not
    
    skip : bool
        Whether or not Sphinx would skip this, given pre-set options
    
    options : object
        Options given to the directive, whose boolean properties are set to True
        if their corresponding flag was given in the directive
    
    
    Returns
    -------
    bool
        True if object should be skipped, False otherwise
    """
    if skip == False:
        if name in exclude_always:
            skip = True
        elif name.startswith("Abstract") or ".Abstract" in name:
            skip = True
        elif name.startswith("_"):
            skip = True
        elif getattr(obj, "plastid_skipdoc", False) == True:
            skip = True
        elif name in exclude_if_no_redoc:
            for cls in exclude_if_no_redoc_base_classes:
                if isinstance(obj, cls):
                    try:
                        base_doc = getattr(cls, name).__doc__
                    except AttributeError:
                        base_doc = ""
                    skip |= obj.__doc__ == base_doc

    return skip


# intersphinx config -----------------------------------------------------------

intersphinx_mapping = {
    "cython"       : ("https://docs.cython.org/en/latest/", None),
    "HTSeq"        : ("https://www-huber.embl.de/users/anders/HTSeq/doc/", None),
    "matplotlib"   : ("https://matplotlib.org/stable/", None),
    "numpy"        : ("https://numpy.org/doc/stable/", None),
    "pandas"       : ("https://pandas-docs.github.io/pandas-docs-travis/", None),
    "pysam"        : ("https://pysam.readthedocs.io/en/latest/", None),
    "python"       : ("https://docs.python.org/3/", None),
    "scipy"        : ("https://docs.scipy.org/doc/scipy/", None),
    "twobitreader" : ("https://pythonhosted.org/twobitreader/", None),
}


# setup custom info ------------------------------------------------------------


def setup(app):
    """Activate custom event handlers for autodoc"""
    app.connect("autodoc-skip-member", autodoc_skip_member)
    app.add_css_file('css/custom.css')


# ------------------------------------------------------------------------------

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The reST default role (used for this markup: `text`) to use for all
# documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False

# -- Options for HTML output ----------------------------------------------

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'plastid-doc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    ('index', 'plastid.tex', u'Plastid Documentation', u'Joshua G. Dunn', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [('index', 'plastid', u'plastid Documentation', [u'Joshua Griffin Dunn'], 1)]

# If true, show URL addresses after external links.
#man_show_urls = False

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        'index', 'plastid', u'plastid Documentation', u'Joshua Griffin Dunn', 'plastid',
        'One line description of project.', 'Miscellaneous'
    ),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False



#!/usr/bin/env python
"""This package contains several object types useful for genomic analyses.

Package overview
================

    =============================================  ==================================================================
    **Submodule**                                   **Description**
    ---------------------------------------------  ------------------------------------------------------------------
    :py:mod:`~plastid.genomics.genome_array`         Array-like objects indexed by chromosome, position, and strand
                                                     
    :py:mod:`~plastid.genomics.genome_hash`          Dictionary-like objects associate features
                                                     with genomic positions, allowing fast query of which
                                                     features overlap a region or nucleotide

    :py:mod:`~plastid.genomics.map_factories`        :term:`Mapping rules <mapping rule>` for
                                                     :class:`~plastid.genomics.genome_array.BAMGenomeArray`
                                                   
    :py:mod:`~plastid.genomics.roitools`             Objects that represent genomic :term:`features <feature>`,
                                                     such as genes or transcripts
    
    :py:mod:`~plastid.genomics.seqtools`             Functions for manipulating nucleic acid
                                                     sequences
                                            
    :py:mod:`~plastid.genomics.splicing`             Functions for manipulating splice junctions
    =============================================  ==================================================================
"""



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.readers.blat`

References
----------
http://pombe.nci.nih.gov/genome/goldenPath/help/blatSpec.html
"""
import unittest
import numpy
from plastid.util.services.mini2to3 import cStringIO
from nose.plugins.attrib import attr
from plastid.genomics.roitools import SegmentChain
from plastid.readers.psl import PSL_Reader, BundledPSL_Reader
from plastid.readers.bed import BED_Reader
from plastid.test.ref_files import MINI


@attr(test="unit")
class TestPSL_Reader(unittest.TestCase):
    def test_iter_finds_all(self):
        with open(MINI["bed_file"]) as fh:
            bed_transcripts = [
                X for X in BED_Reader(fh, return_type=SegmentChain) if "intron" not in X.get_name()
            ]
            bed_transcripts = [X for X in bed_transcripts if "repeat" not in X.get_name()]
        with open(MINI["psl_file"]) as fh:
            psl_transcripts = [X for X in PSL_Reader(fh) if "repeat" not in X.get_name()]

        self.assertEqual(
            len(bed_transcripts), len(psl_transcripts),
            "Length mismatch: %s vs %s" % (len(bed_transcripts), len(psl_transcripts))
        )
        self.assertGreater(len(bed_transcripts), 0)

    def test_iter_reads_correct(self):
        with open(MINI["bed_file"]) as fh:
            bed_transcripts = [
                X for X in BED_Reader(fh, return_type=SegmentChain) if "intron" not in X.get_name()
            ]
            bed_transcripts = [X for X in bed_transcripts if "repeat" not in X.get_name()]
        with open(MINI["psl_file"]) as fh:
            psl_transcripts = [X for X in PSL_Reader(fh) if "repeat" not in X.get_name()]
        for bed, psl in zip(bed_transcripts, psl_transcripts):
            bed_name = bed.get_name()
            psl_name = psl.get_name()
            self.assertEqual(bed_name, psl_name, "Name mismatch: %s vs %s" % (bed_name, psl_name))
            psl_positions = psl.get_position_set()
            bed_positions = bed.get_position_set()
            self.assertEqual(
                bed_positions, psl_positions,
                "Position mismatch: %s vs %s" % (bed_positions, psl_positions)
            )
            bed_chrom = bed.spanning_segment.chrom
            psl_chrom = psl.spanning_segment.chrom
            self.assertEqual(
                bed_chrom, psl_chrom, "Chromosome mismatch: %s vs %s" % (bed_chrom, psl_chrom)
            )
            bed_strand = bed.spanning_segment.strand
            psl_strand = psl.spanning_segment.strand
            self.assertEqual(
                bed_strand, psl_strand, "Strand mismatch: %s vs %s" % (bed_strand, psl_strand)
            )


@attr(test="unit")
class TestBundledPSL_Reader(unittest.TestCase):
    def test_iter_bundles(self):
        # create a list of entries with same query name, so that these will be bundled by reader
        with open(MINI["psl_file"]) as fh:
            psl_transcripts = list(PSL_Reader(fh))

        repeats = numpy.random.randint(1, high=10, size=len(psl_transcripts))
        stmp = ""
        for tx, r in zip(psl_transcripts, repeats):
            for _ in range(r):
                stmp += tx.as_psl()

        fakefile = cStringIO.StringIO(stmp)
        reader = BundledPSL_Reader(fakefile)
        for n, tx_group in enumerate(reader):
            # make sure groups are correct length
            self.assertEqual(len(tx_group), repeats[n])

            # make sure each group has the correct query name
            names = [X.get_name() for X in tx_group]
            self.assertEqual(len(set(names)), 1)
            self.assertEqual(names[0], psl_transcripts[n].get_name())

        self.assertEqual(n + 1, len(repeats))
        self.assertEqual(n + 1, len(psl_transcripts))



#!/usr/bin/env python
"""Utilities for mutating and searching nucleic acid sequences.

Contents
--------

.. autosummary::

   TwoBitSeqRecordAdaptor
   mutate_seqs
   seq_to_regex
   IUPAC_TABLE
"""
import random, re
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from twobitreader import TwoBitFile

IUPAC_TABLE = {
    "A": "A",
    "C": "C",
    "T": ("T", "U"),
    "U": ("T", "U"),
    "G": "G",
    "N": ("A", "C", "T", "G", "U"),

    # puRines
    "R": ("A", "G"),

    # pYrimidines
    "Y": ("C", "T", "U"),

    # Strong binding
    "S": ("G", "C"),

    # Weak binding
    "W": ("A", "T", "U"),

     # Keto-functionalized
    "K": ("G", "T", "U"),

    # aMino-functionalized
    "M": ("A", "C"),

    # B = any but A
    "B": ("C", "G", "T", "U"),

    # C = any but C
    "D": ("A", "G", "T", "U"),

    # H = any but G
    "H": ("A", "C", "T", "U"),

    # V = any but T or U
    "V": ("A", "C", "G"),
}
"""Dictionary mapping IUPAC nucleotide symbols to tuples of nucleotides
they represent (e.g. R -> (A, G) )
"""


def seq_to_regex(inp, flags=0):
    """Convert a nucleotide sequence of IUPAC nucleotide characters as a regular expression.
    Ambiguous IUPAC characters are converted to groups (e.g. `'Y'` to `'[CTU]'`),
    and T and U are considered equivalent.
    
    Parameters
    ----------
    inp : str
        Nucleotide sequence using IUPAC nucleotide codes
    
    flags : int, optional
        Flags to pass to :py:func:`re.compile` (Default: 0 / no flags)
        
    Examples
    --------
    Convert a sequence to a regex::
    
        >>> seq_to_regex("CARYYA").pattern
        'CA[AG][CTU][CTU]A'
    
    
    Returns
    -------
    :py:class:`re.RegexObject`
        Regular expression pattern corresponding to IUPAC sequence in `inp`
    """
    out = []
    for ch in inp:
        if len(IUPAC_TABLE.get(ch, ch)) == 1:
            out.append(ch)
        else:
            out.append("[" + "".join(IUPAC_TABLE.get(ch, ch)) + "]")

    return re.compile("".join(out), flags=flags)


def mutate_seqs(seqs, nucleotides="NACTG", mutations=1):
    """Generate all sequences within `mutations` distance from a reference sequence
    
    Parameters
    ----------
    seqs : str or list of str
        Single reference sequence (a string) or a group of strings
                        
    nucleotides : list of char, optional
        Permitted nucleotide substitutions (Default: `'NACTG'`)
    
    mutations : int, optional
        Number of substitutions to make (Default: `1`)
    
    
    Returns
    -------
    set
        all sequences within `mutations` substitutions from the sequence(s)
        specified in `seqs`
    """
    if isinstance(seqs, str):
        seqs = [seqs]
    if mutations == 0:
        return set(seqs)
    else:
        seqsout = []
        for seq in seqs:
            for nuc in nucleotides:
                for i in range(len(seq)):
                    newseq = list(seq)[:]
                    newseq[i] = nuc
                    seqsout.append("".join(newseq))

        seqsout.extend(mutate_seqs(seqsout, nucleotides=nucleotides, mutations=mutations - 1))
        return set(seqsout) | set(seqs)


def random_seq(size, nucleotides="ACTG"):
    """Generate a random nucleotide sequence of length `size` and composition `nucleotides`
    
    Parameters
    ----------
    size : int
        length of desired sequence
        
    nucleotides : str, optional
        string of nucleotides to use in sequence, in desired base composition
        (i.e. need not be unique; can supply `'AATCG'` to increase `'A'` bias.
        Default: `'ACTG'`)
        
    Returns
    -------
    str : randomized DNA sequence
    """
    seq = "".join([nucleotides[random.randrange(0, len(nucleotides))] for _ in range(0, size)])
    return seq


def revive(twobitreader, seqname):
    _TwoBitSeqProxy(twobitreader[seqname], twobitreader)


class _TwoBitSeqProxy(object):
    """Adaptor class that fetches :class:`Bio.SeqRecord.SeqRecord`
    objects from :class:`twobitreader.TwoBitSequence` objects

    Defines `seq` property and :meth:`~_TwoBitSeqProxy.reverse_complement` method
    """

    def __init__(self, twobitfile, key):
        """
        Parameters
        ----------
        twobitseq : :class:`twobitfile.TwoBitSequence`
        """
        self.name = key
        self._twobitfile = twobitfile
        self.twobitseq = twobitfile[key]
        self._seq = None

    def __reduce__(self):
        return (_TwoBitSeqProxy, (self._twobitfile, self.name))

    def __getitem__(self, slice_):
        return SeqRecord(
            Seq(self.twobitseq.get_slice(min_=slice_.start, max_=slice_.stop))
        )

    def __len__(self):
        return len(self.twobitseq)

    def __str__(self):
        """Return sequence in `self.twobitseq` as str"""
        return str(self.twobitseq)

    def __getattr__(self, attr):
        if attr == "seq":
            if self._seq is None:
                self._seq = Seq(str(self.twobitseq))

            return self._seq

    def reverse_complement(self):
        """Return the reverse complement of the TwoBitSequence

        Returns
        -------
        :class:`Bio.SeqRecord.SeqRecord`
            Reverse complement of the sequence held in `self.twobitseq`
        """
        return SeqRecord(self.seq).reverse_complement()


class TwoBitSeqRecordAdaptor(object):
    """Adaptor class that makes a :class:`twobitreader.TwoBitFile` behave
    like a dictionary of :class:`Bio.SeqRecord.SeqRecord` objects.
    """

    def __init__(self, fh):
        self.twobitfile = TwoBitFile(fh)
        self._filename = fh
        self._chroms = {K: _TwoBitSeqProxy(self.twobitfile, K) for K in self.twobitfile}

    def __reduce__(self):
        return (TwoBitSeqRecordAdaptor, (self._filename, ))

    def __getitem__(self, key):
        return self._chroms[key]

    def __getattr__(self, attr):
        try:
            return getattr(self._chroms, attr)
        except AttributeError:
            return getattr(self.twobitfile, attr)

    def __iter__(self):
        return iter(self._chroms)

    def __len__(self):
        return len(self._chroms)



#!/usr/bin/env python
"""This script empirically determines which positions in a genome yield 
multimapping reads under a given set of alignment parameters. These positions
are saved in a `BED`_-formatted :term:`mask file`, so that they may be excluded
as from further analyses.

To identify multimapping regions, a genome sequence is diced into :term:`k-mers
<k-mer>` and the :term:`k-mers` aligned back to the genome. Positions in the
genome that give rise to :term:`k-mers <k-mer>` that align equally well to more
than one genomic location are then marked as multimapping..

`k` is specified by the user, as are the alignment parameters.


Output files
------------
The following files are made:

    OUTBASE_READLENGTH_MISMATCHES_crossmap.bed
        Final :term:`mask file` annotation, in `BED`_ format

    OUTBASE_READLENGTH_MISMATCHES_CHROMOSOME_kmers.fa
        :term:`K-mers <k-mer>` derived from chromosome `CHROMOSOME`. These
        files can be reused in subsequent runs allowing a different number of
        mismatches, using the ``--have_kmers`` option

where:

  - `OUTBASE` is a name meaningful to the user
  - `READLENGTH` is the :term:`k-mer` length chosen by the user
  - `MISMATCHES` is the number of mismatches permitted during alignment,
    also set by the user.



Considerations for large genomes
--------------------------------

For large genomes (e.g. vertebrate, plant, or some *very* big amoebas):

  - |crossmap| can require a ton of memory if genome sequence is stored 
    in a fasta file. If |crossmap| maxes out your system's memory, it may
    be terminated by your system before it completes.

    Consider converting the file to a `2bit`_ file to save memory and
    avoid this potential problem

  - Enabling mismatches with short read sizes will make |crossmap| take
    a lot longer, especially on large genomes. Consider using ``--mismatches
    0`` if you run into this problem 

  - Using multiple processes (e.g. via ``-p 2``) will speed |crossmap|'s
    execution time, but will increase its memory footprint, as each process
    will need its own memory space to create and align k-mers from chromosomal
    sequence

  - By default, |crossmap| creates `BED`_ files. Consider converting these to
    `BigBed`_ files will save substantial amounts of time and memory in the future.

    This can be achieved using Jim Kent's ``bedToBigBed`` utility as follows
    (from the terminal)::

        $ bowtie-inspect --summary BOWTIE_INDEX | grep Sequence |\\
                         cut -f2,3 | sed -e "s/\([^ ]\+\).*\\t/\\1\\t/"  >OUTFILE.sizes
        $ sort -k1,1 -k2,2n OUTBASE.bed > OUTBASE_sorted.bed
        $ bedToBigBed OUTBASE_sorted.bed OUTBASE.sizes OUTBASE_sorted.bb


    See https://github.com/ENCODE-DCC/kentUtils/tree/master/src/product/scripts
    for download & documentation of Kent utilities
"""
__author__ = "joshua"
import argparse
import functools
import glob
import inspect
import multiprocessing
import os
import re
import shutil
import subprocess
import sys
from plastid.util.io.filters import NameDateWriter, AbstractReader
from plastid.util.io.openers import get_short_name, argsopener
from plastid.genomics.roitools import SegmentChain, positionlist_to_segments, GenomicSegment
from plastid.util.scriptlib.help_formatters import format_module_docstring
from plastid.util.services.mini2to3 import xrange
from plastid.util.services.exceptions import MalformedFileError
from plastid.util.scriptlib.argparsers import SequenceParser, BaseParser


namepat = re.compile(r"(.*):([0-9]+)\(\+\)")
printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))

BigBedMessage = """Crossmap complete and saved as 'OUTFILE.bed'.

    For large (e.g. mammalian) genomes, it is highly recommended to convert
    the BED-format output to a BigBed file, using Jim Kent's bedToBigBed
    utility as follows (from the terminal):

        $ bowtie-inspect --summary BOWTIE_INDEX | grep Sequence |\\
                         cut -f2,3 | sed -e "s/\\([^ ]\+\\).*\\t/\\1\\t/"  >OUTFILE.sizes
        $ sort -k1,1 -k2,2n OUTFILE.bed > OUTFILE_sorted.bed
        $ bedToBigBed OUTFILE_sorted.bed OUTFILE.sizes OUTFILE_sorted.bb


    See https://github.com/ENCODE-DCC/kentUtils/tree/master/src/product/scripts
    for download & documentation of Kent utilities


"""


def simulate_reads(seq_record, fh=sys.stdout, k=30):
    """Chops a DNA sequence into :term:`k-mers <k-mer>`, mimicking a sequencing
    run.  Output is delivered in fasta format. Sequences are named for position
    of origin using 0-based indices.

    Parameters
    ----------
    seq_record : :py:class:`Bio.SeqRecord.SeqRecord`
        DNA sequence

    fh : file-like
        filehandle to write output 

    k : int, optional
        length of k-mers to generate (Default: `30`)
    """
    seq = str(seq_record.seq)

    for x in xrange(0, len(seq_record) - k + 1):
        fh.write(">%s:%s(+)\n" % (seq_record.name, x))
        fh.write("%s\n" % seq[x:x + k])

    return None


class FastaNameReader(AbstractReader):
    """Returns names of sequences in a fasta file"""

    def filter(self, line):
        """Return next sequence name in a fasta file

        Parameters
        ----------
        line : str
            Line of text

        Returns
        -------
        str
            Name of next sequence, excluding prefix `'>'` and line terminator
        """
        if line.startswith(">"):
            return line[1:].rstrip()
        else:
            return self.__next__()


def revcomp_mask_chain(seg, k, offset=0):
    """Reverse-complement a single-interval mask, correcting for `offset`.
    
    Parameters
    ----------
    seg : |SegmentChain|
        Plus-strand mask, including `offset`

    k : int
        Length of k-mers

    offset : int, optional
        Offset from 5' end of read at which to map mask (Default: `0`)

    Returns
    -------
    |SegmentChain|
        Mask on minus strand corresponding to `seg`
    """
    # Algorithm note:
    #
    #     Let
    #         FW = plus-strand coordinate
    #         RC = minus-strand coordinate
    #
    #     Then
    #         RC = FW + k - 1 - offset
    #
    #     But we are given FW + offset, so:
    #
    #         RC + offset = (FW + offset) + k - 1 - offset
    #         RC = (FW + offset) + k - 1 - 2*offset
    span = seg.spanning_segment
    new_offset = k - 1 - 2 * offset
    ivminus = GenomicSegment(span.chrom, span.start + new_offset, span.end + new_offset, "-")
    return SegmentChain(ivminus)


def fa_to_bed(toomany_fh, k, offset=0):
    """Create a `BED`_ file indicating genomic origins of reads in a `bowtie`_
    ``toomany`` file

    Parameters
    ----------
    toomany_fh : file-like
        Open filehandle to fasta-formatted ``toomany`` file from `bowtie`_

    k : int
        Length of k-mers

    offset : int, optional
        Offset from 5' end of read at which to map read, if any (Default: `0`)

    Yields
    ------
    |SegmentChain|
        Plus-strand |SegmentChain| representing a repetitive region

    |SegmentChain|
        Minus-strand |SegmentChain| representing a repetitive region
    """
    last_chrom = chrom = None
    last_pos = None
    start_pos = None
    reader = FastaNameReader(toomany_fh)

    for n, read_name in enumerate(reader):
        chrom, pos = namepat.search(read_name).groups()
        pos = int(pos) + offset
        if chrom != last_chrom:
            if last_chrom is not None:
                plus_chain = SegmentChain(GenomicSegment(last_chrom, start_pos, last_pos + 1, "+"))
                minus_chain = revcomp_mask_chain(plus_chain, k, offset)
                last_chrom = chrom
                start_pos = pos
                last_pos = pos
                yield plus_chain, minus_chain
            else:
                last_chrom = chrom
                start_pos = pos
                last_pos = pos
        else:
            delta = pos - last_pos
            if delta > 1:
                plus_chain = SegmentChain(GenomicSegment(chrom, start_pos, last_pos + 1, "+"))
                minus_chain = revcomp_mask_chain(plus_chain, k, offset)
                last_pos = pos
                start_pos = pos
                yield plus_chain, minus_chain
            elif delta == 1:
                last_pos = pos
            else:
                msg = "k-mers are not sorted at read '%s' ! Aborting." % read_name
                raise MalformedFileError(toomany_fh, msg, line_num=n)

    if chrom is not None:
        plus_chain = SegmentChain(GenomicSegment(chrom, start_pos, last_pos + 1, "+"))
        minus_chain = revcomp_mask_chain(plus_chain, k, offset)
        yield plus_chain, minus_chain


def chrom_worker(chrom_seq, args=None):
    name, seq_or_kmers = chrom_seq
    printer.write("Processing chromosome '%s' ..." % name)
    base = "%s_%s_%s_%s" % (args.outbase, args.read_length, args.mismatches, name)
    kmer_file = "%s_kmers.fa" % base
    toomany_file = "%s_multimap.fa" % base
    bed_file = "%s_crossmap.bed" % base

    if args.have_kmers == False:
        with open(kmer_file, "w") as kmer:
            simulate_reads(seq_or_kmers, kmer, args.read_length)

        kmer.close()
    else:
        kmer_file = seq_or_kmers

    argdict = {
        "mismatches": args.mismatches,
        "processors": 1,
        "bowtie"    : args.bowtie,
        "toomany"   : toomany_file,
        "kmers"     : kmer_file,
        "ebwt"      : args.ebwt,
        "null"      : os.devnull,
    } # yapf: disable

    cmd = "%(bowtie)s -m1 -a --best -f -v %(mismatches)s -p %(processors)s %(ebwt)s %(kmers)s --max %(toomany)s >%(null)s" % argdict

    printer.write("Aligning %s-mers for chromosome '%s' :\n\t'%s'" % (args.read_length, name, cmd))
    try:
        retcode = subprocess.call(cmd, shell=True)
        if retcode < 0 or retcode == 2:
            printer.write(
                "Alignment for chromosome '%s' terminated with status %s" % (name, retcode)
            )
        else:
            if os.path.exists(toomany_file):
                printer.write(
                    "Assembling multimappers from chromosome '%s' into crossmap ..." % name
                )
                with argsopener(bed_file, args, "w") as bed_out:
                    for plus_chain, minus_chain in fa_to_bed(open(toomany_file), args.read_length,
                                                             offset=args.offset):
                        bed_out.write(plus_chain.as_bed())
                        bed_out.write(minus_chain.as_bed())

                    bed_out.close()
                os.remove(toomany_file)

            else:
                printer.write(
                    "Could not find multimapper source file '%s' ." % toomany_file
                )

    except OSError as e:
        printer.write("Alignment failed for chromosome '%s': %s" % (name, e))

    printer.write("Cleaning up chromosome '%s' ..." % name)
    if args.have_kmers == False and args.save_kmers == False:
        os.remove(kmer_file)

    return bed_file


def main(argv=sys.argv[1:]):
    """Command-line program

    Parameters
    ----------
	argv : list, optional
		A list of command-line arguments, which will be processed
		as if the script were called from the command line if
		:py:func:`main` is called directly.

        Default: `sys.argv[1:]`. The command-line arguments, if the script is
        invoked from the command line
    """
    sp = SequenceParser()
    bp = BaseParser()

    parser = argparse.ArgumentParser(
        description     = format_module_docstring(__doc__),
        formatter_class = argparse.RawDescriptionHelpFormatter,
        parents         = [bp.get_parser(), sp.get_parser()],
    )
    parser.add_argument(
        "-k",
        dest    = "read_length",
        metavar = "READ_LENGTH",
        type    = int,
        default = 29,
        help    = "K-mer length to generate from input file. (Default: 29)",
    )
    parser.add_argument(
        "--offset",
        type    = int,
        default = 14,
        help    =
        "Offset from 5' end of plus-strand read at which to attribute score (Default: 14)"
    )
    parser.add_argument(
        "--mismatches",
        metavar="N",
        type=int,
        default=0,
        help="Number of mismatches tolerated in alignment. (Default: 0)"
    )
    parser.add_argument(
        "--bowtie",
        dest    = "bowtie",
        default = "/usr/bin/bowtie",
        type    = str,
        help    = "Location of bowtie binary (Default: /usr/bin/bowtie)",
    )
    parser.add_argument(
        "--have_kmers",
        default=False,
        action="store_true",
        help=(
            "If specified,  use k-mer files from previous run. In this case "
            "'sequence_file' should be the value 'outbase' from the k-mer "
            "files you want to use. (Default: False)"
        ),
    )
    parser.add_argument(
        "--save_kmers",
        default = False,
        action  = "store_true",
        help    = "Save k-mer files for reuse in a subsequent run.",
    )
    parser.add_argument(
        "-p",
        "--processes",
        type    = int,
        default = 2,
        metavar = "N",
        help    = (
            "Number of processes to use (should be <= number of chromosomes; "
            "default: 2)"
        ),
    )
    parser.add_argument(
        "ebwt",
        type = str,
        help = (
            "Bowtie index of genome against which crossmap will be made. "
            "In most cases, should be generated from the same sequences that "
            "are in `sequence_file`."
        ),
    )
    parser.add_argument("outbase", type=str, help="Basename for output files")

    args = parser.parse_args(argv)
    bp.get_base_ops_from_args(args)

    #filenames
    base = "%s_%s_%s" % (args.outbase, args.read_length, args.mismatches)
    bed_file = "%s_crossmap.bed" % base

    if args.have_kmers == True:
        kmer_files = glob.glob(args.sequence_file + "*kmers.fa")
        seq_pat = re.compile(r".*_([^_]*)_kmers.fa")
        seqs = {seq_pat.search(X).groups()[0]: X for X in kmer_files}
    else:
        seqs = sp.get_seqdict_from_args(args, index=True)

    worker = functools.partial(chrom_worker, args=args)
    chroms = seqs.items()

    pool = multiprocessing.Pool(processes=args.processes)
    bed_filenames = pool.map(worker, chroms, 1)
    pool.close()
    pool.join()

    with open(bed_file, "w") as fout:
        for f in sorted(bed_filenames):
            if os.path.exists(f):
                shutil.copyfileobj(open(f, "r"), fout)
                os.remove(f)
            else:
                printer.write("Could not find intermediate bed file '%s' ." % f)

    fout.close()

    printer.write("Done.")
    printer.write(
        BigBedMessage.replace(
            "OUTFILE",
            bed_file.replace(".bed", "")
        ).replace("BOWTIE_INDEX", args.ebwt)
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Test suites for :py:mod:`plastid.readers.autosql`"""
import unittest
import itertools
import warnings
import nose.tools
from nose.plugins.attrib import attr
from collections import OrderedDict
from plastid.util.services.decorators import skip_if_abstract
from plastid.readers.autosql import AutoSqlField, SizedAutoSqlField, ValuesAutoSqlField, AutoSqlDeclaration, AbstractAutoSqlElement
from plastid.util.services.exceptions import pl_once_registry

_autosql_declarations = [
    """table easy_table
"A table with a comment on the next line" 
    (
    uint number auto; "a number with a token"
    uint [3] points ; "r,g,b values"
    lstring  my_string ; "a long string"
    uint a_field_size ; "the size for the next field"
    float [a_field_size] float_array ; "an array of floats"
    set(a,b,c) alpha ; "the first three letters of the alphabet"
    )
""", """table easy_table "A table with a comment on the next line" 
    (
    uint number auto; "a number with a token"
    uint [3] points ; "r,g,b values"
    lstring  my_string ; "a long string"
    uint a_field_size ; "the size for the next field"
    float [a_field_size] float_array ; "an array of floats"
    set(a,b,c) alpha ; "the first three letters of the alphabet"
    )
""", """table easy_table "A table with a comment on the next line" ( 
uint number auto; "a number with a token"
uint [3] points ; "r,g,b values"
lstring  my_string ; "a long string"
uint a_field_size ; "the size for the next field"
float [a_field_size] float_array ; "an array of floats"
set(a,b,c) alpha ; "the first three letters of the alphabet")
"""
]

match_texts = {
    "AutoSqlField": [
        'uint red ; "red rgb value"',
        'uint red;  "red rgb value"',
        '    uint red ;  "red rgb value"',
        'uint red;  "red rgb value"    ',
        'float orange ; "orange value"',
        'string  desc    ; "description"',
        'lstring  desc    ; "description"',
        'float orange primary ; "orange value"',
        'float orange primary index ; "orange value"',
        'float orange primary index[5] ; "orange value"',
        'float orange primary index [5] ; "orange value"',
        'float orange primary index [5] auto ; "orange value"',
        'float orange primary index[5] auto ; "orange value"',
        'float orange primary index [ 5 ] auto ; "orange value"',
        'float orange auto primary index [5] ; "orange value"',
        'float orange auto primary index; "orange value"',
        'float orange primary auto index; "orange value"',
        'float orange index primary auto; "orange value"',
        'float orange index [ 5] primary auto; "orange value"',
    ],
    "SizedAutoSqlField": [
        'uint [ 3 ] points ; "r,g,b values"',
        'uint [3] points ; "r,g,b values"',
        'uint[3] points ; "r,g,b values"',
        'uint[3] points; "r,g,b values"',
        'float [ 2 ] cmplx ; "real and imaginary parts"',
        'string [10] text ; "some text"',
        'uint [ numpoints ] points ; "r,g,b values"',
        'uint [ 3 ] points primary ; "r,g,b values"',
        'uint [ 3 ] points index ; "r,g,b values"',
        'uint [ 3 ] points auto ; "r,g,b values"',
        'uint [ 3 ] points primary index auto ; "r,g,b values"',
        'uint [ 3 ] points index auto primary ; "r,g,b values"',
        'uint [ 3 ] points primary index auto; "r,g,b values"',
        'uint [ 3 ] points primary auto ; "r,g,b values"',
        'uint [ 3 ] points index primary ; "r,g,b values"',
        'uint [ 3 ] points primary index [12] auto ; "r,g,b values"',
        'uint [ 3 ] points auto primary index [12]; "r,g,b values"',
    ],
    "ValuesAutoSqlField": [
        'set (a,b,c) alphabet ; "the first three letters"',
        'set ( a , b , c ) alphabet ; "the first three letters"',
        'set ( a , b , c ) alphabet; "the first three letters"',
        'set (a, b, c) alphabet ; "the first three letters"',
        'enum (a,b,c) alphabet ; "the first three letters"',
        'enum ( a , b , c ) alphabet ; "the first three letters"',
        'enum ( a , b , c ) alphabet; "the first three letters"',
        'enum (a, b, c) alphabet ; "the first three letters"',
        'set (a,b,c) alphabet primary ; "the first three letters"',
        'set (a,b,c) alphabet index ; "the first three letters"',
        'set (a,b,c) alphabet primary index [2] ; "the first three letters"',
        'set (a,b,c) alphabet index [2] primary ; "the first three letters"',
    ],
    "AutoSqlDeclaration":
    _autosql_declarations,
}
"""Text snippets that should be matched by the RegExes in various classes"""


@attr(test="unit")
def test_mask_comments():
    # test comment masking on a variety of fields and on a declaration
    # make sure strings are returned as expected, and make sure location
    # of comment in fields matches index of first quotation mark
    field_tests = [
        'uint red ; "red rgb value"',
        'uint red;  "red rgb value"',
        '    uint red ;  "red rgb value"',
        'float orange primary index[5] ; "orange value"',
        'float orange primary index [5] ; "orange value"',
        'float orange primary index [5] auto ; "orange value"',
        'float orange primary index[5] auto ; "orange value"',
        'enum ( a , b , c ) alphabet ; "the first three letters"',
        'enum ( a , b , c ) alphabet; "the first three letters"',
        'enum (a, b, c) alphabet ; "the first three letters"',
        'set (a,b,c) alphabet primary ; "the first three letters"',
        'set (a,b,c) alphabet index ; "the first three letters"',
        'set (a,b,c) alphabet primary index [2] ; "the first three letters"',
        'uint[3] points ; "r,g,b values"',
        'uint[3] points; "r,g,b values"',
        'float [ 2 ] cmplx ; "real and imaginary parts"',
        'string [10] text ; "some text"',
        'uint [ numpoints ] points ; "r,g,b values"',
        'uint [ 3 ] points primary ; "r,g,b values"',
        'uint [ 3 ] points index ; "r,g,b values"',
        'uint [ 3 ] points auto ; "r,g,b values"',
    ]
    field_test_results = [
        'uint red ; "xxxxxxxxxxxxx"',
        'uint red;  "xxxxxxxxxxxxx"',
        '    uint red ;  "xxxxxxxxxxxxx"',
        'float orange primary index[5] ; "xxxxxxxxxxxx"',
        'float orange primary index [5] ; "xxxxxxxxxxxx"',
        'float orange primary index [5] auto ; "xxxxxxxxxxxx"',
        'float orange primary index[5] auto ; "xxxxxxxxxxxx"',
        'enum ( a , b , c ) alphabet ; "xxxxxxxxxxxxxxxxxxxxxxx"',
        'enum ( a , b , c ) alphabet; "xxxxxxxxxxxxxxxxxxxxxxx"',
        'enum (a, b, c) alphabet ; "xxxxxxxxxxxxxxxxxxxxxxx"',
        'set (a,b,c) alphabet primary ; "xxxxxxxxxxxxxxxxxxxxxxx"',
        'set (a,b,c) alphabet index ; "xxxxxxxxxxxxxxxxxxxxxxx"',
        'set (a,b,c) alphabet primary index [2] ; "xxxxxxxxxxxxxxxxxxxxxxx"',
        'uint[3] points ; "xxxxxxxxxxxx"',
        'uint[3] points; "xxxxxxxxxxxx"',
        'float [ 2 ] cmplx ; "xxxxxxxxxxxxxxxxxxxxxxxx"',
        'string [10] text ; "xxxxxxxxx"',
        'uint [ numpoints ] points ; "xxxxxxxxxxxx"',
        'uint [ 3 ] points primary ; "xxxxxxxxxxxx"',
        'uint [ 3 ] points index ; "xxxxxxxxxxxx"',
        'uint [ 3 ] points auto ; "xxxxxxxxxxxx"',
    ]

    declaration_test = 'table my_table "a test table for me " (\n' + "\n".join(field_tests) + "\n)"
    declaration_expected = 'table my_table "xxxxxxxxxxxxxxxxxxxx" (\n' + "\n".join(
        field_test_results
    ) + "\n)"
    for test, expected in zip(field_tests, field_test_results):
        found = AbstractAutoSqlElement.mask_comments(test)
        expected_start = test.index('"')
        nose.tools.assert_equal(
            expected, found[0], "Failed to mask field comment '%s'. Expected '%s', got '%s'" %
            (test, expected, found[0])
        )
        nose.tools.assert_equal(expected_start + 1, found[1][0][0])

    declaration_found = AbstractAutoSqlElement.mask_comments(declaration_test)
    nose.tools.assert_equal(
        declaration_expected, declaration_found[0],
        "Failed to mask comment '%s'.Expected:\n'%s'\n Got:\n'%s'" %
        (declaration_test, declaration_expected, declaration_found[0])
    )


@attr(test="unit")
class TestAbstractAutoSqlElement(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # define all of these in subclasses
        cls.test_class = None

        # instances of autoSql text that match cls.test_class.match_pattern
        cls.matching_text = []

        # instances of autoSql text that do not match cls.test_class.match_pattern
        cls.non_matching_text = []

        # dictionary mapping indices of autoSql specs in cls.matching_text
        # to a list of tuples of unparsed text and expected result objects
        cls.expected_values = {}

        # record holding values
        cls.test_record = OrderedDict()

    @skip_if_abstract
    def test_match_pattern(self):
        for c, text in enumerate(self.matching_text):
            self.assertTrue(
                self.test_class.match_pattern.search(text) is not None,
                "%s failed to match text '%s'" % (self.test_class.__name__, text)
            )

        self.assertGreater(
            c + 1, 0, "%s no matches tested in test_match_pattern" % self.test_class.__name__
        )

    @skip_if_abstract
    def test_not_match_pattern(self):
        for c, text in enumerate(self.non_matching_text):
            self.assertFalse(
                self.test_class.match_pattern.search(text) is not None,
                "%s matched text that should have been unmatchable: %s" %
                (self.test_class.__name__, text)
            )

        self.assertGreater(
            c + 1, 0, "%s no matches tested in test_not_match_pattern" % self.test_class.__name__
        )

    @skip_if_abstract
    def test_add_type(self):
        my_instance = self.test_class(self.matching_text[0])
        fname = "new_type"
        self.assertNotIn(fname, my_instance.field_types)
        my_instance.add_type(fname, int)
        self.assertIn(fname, my_instance.field_types)
        self.assertEqual(my_instance.field_types[fname], int)

    @skip_if_abstract
    def test_call(self):
        for n, declaration in enumerate(self.matching_text):
            parser = self.test_class(declaration)
            for test_str, expected in self.expected_values[n]:
                found = parser(test_str, self.test_record)
                self.assertEqual(
                    found, expected, "%s parser results incorrect. Got '%s', expected '%s'" %
                    (self.test_class.__name__, found, expected)
                )


@attr(test="unit")
class TestAutoSqlField(TestAbstractAutoSqlElement):
    @classmethod
    def setUpClass(cls):
        # define all of these in subclasses
        cls.test_class = AutoSqlField
        cls.match_key = cls.test_class.__name__

        # instances of autoSql text that match cls.test_class.match_pattern
        cls.matching_text = match_texts[cls.match_key]

        # instances of autoSql text that do not match cls.test_class.match_pattern
        cls.non_matching_text = list(
            itertools.chain(
                *map(lambda x: x[1], filter(lambda x: x[0] != cls.match_key, match_texts.items()))
            )
        )

        cls.test_record = OrderedDict()

        # list of lists of tuples of unparsed text and expected result objects
        # for each autoSql declaration in cls.matching_text
        cls.expected_values = [
            [
                ("3", 3),
                (" 0", 0),
                ("20", 20),
            ],
            [
                ("3", 3),
                (" 0", 0),
                ("20", 20),
            ],
            [
                ("3", 3),
                (" 0", 0),
                ("20", 20),
            ],
            [
                ("3", 3),
                (" 0", 0),
                ("20", 20),
            ],
            [("3.0", 3.0), (" 0.0", 0.0), ("-275.4", -275.4)],
            [
                ("some text with spaces", "some text with spaces"),
                ("some text,with,commas", "some text,with,commas"),
                ("asf32Faf!#%#@%", "asf32Faf!#%#@%")
            ],
            [
                ("some text with spaces", "some text with spaces"),
                ("some text,with,commas", "some text,with,commas"),
                ("asf32Faf!#%#@%", "asf32Faf!#%#@%")
            ],
        ] + ([[("3.0", 3.0), (" 0.0", 0.0), ("-275.4", -275.4)]] * 12)


@attr(test="unit")
class TestSizedAutoSqlField(TestAbstractAutoSqlElement):
    @classmethod
    def setUpClass(cls):
        # define all of these in subclasses
        cls.test_class = SizedAutoSqlField
        cls.match_key = cls.test_class.__name__

        # instances of autoSql text that match cls.test_class.match_pattern
        cls.matching_text = match_texts[cls.match_key]

        # instances of autoSql text that do not match cls.test_class.match_pattern
        cls.non_matching_text = list(
            itertools.chain(
                *map(lambda x: x[1], filter(lambda x: x[0] != cls.match_key, match_texts.items()))
            )
        )

        cls.test_record = OrderedDict(numpoints=3)

        # list of lists of tuples of unparsed text and expected result objects
        # for each autoSql declaration in cls.matching_text
        cls.expected_values = [
            [
                ("3,4,5", (3, 4, 5)),
                ("3, 4, 5", (3, 4, 5)),
                ("   201,0, 476", (201, 0, 476)),
            ],
            [
                ("3,4,5", (3, 4, 5)),
                ("3, 4, 5", (3, 4, 5)),
                ("   201,0, 476", (201, 0, 476)),
            ],
            [
                ("3,4,5", (3, 4, 5)),
                ("3, 4, 5", (3, 4, 5)),
                ("   201,0, 476", (201, 0, 476)),
            ],
            [
                ("3,4,5", (3, 4, 5)),
                ("3, 4, 5", (3, 4, 5)),
                ("   201,0, 476", (201, 0, 476)),
            ],
            [
                ("5.4,3.2", (5.4, 3.2)),
                ("5.4,3.2   ", (5.4, 3.2)),
                ("    5.4,3.2", (5.4, 3.2)),
                ("5.4,   3.2", (5.4, 3.2)),
            ],
            [("abcdefghij", "abcdefghij")],
            [
                ("3,4,5", (3, 4, 5)),
                ("3, 4, 5", (3, 4, 5)),
                ("   201,0, 476", (201, 0, 476)),
            ],
        ] + (
            [[
                ("3,4,5", (3, 4, 5)),
                ("3, 4, 5", (3, 4, 5)),
                ("   201,0, 476", (201, 0, 476)),
            ]] * 10
        )


@attr(test="unit")
class TestValuesAutoSqlField(TestAbstractAutoSqlElement):
    @classmethod
    def setUpClass(cls):
        # define all of these in subclasses
        cls.test_class = ValuesAutoSqlField
        cls.match_key = cls.test_class.__name__

        # instances of autoSql text that match cls.test_class.match_pattern
        cls.matching_text = match_texts[cls.match_key]

        # instances of autoSql text that do not match cls.test_class.match_pattern
        cls.non_matching_text = list(
            itertools.chain(
                *map(lambda x: x[1], filter(lambda x: x[0] != cls.match_key, match_texts.items()))
            )
        )

        cls.test_record = OrderedDict(r=255, g=5, c=10)

        # list of lists of tuples of unparsed text and expected result objects
        # for each autoSql declaration in cls.matching_text
        cls.expected_values = [
            [
                (("a,b,c"), set(list("abc"))),
                (("a,c,b"), set(list("abc"))),
                (("a,b"), set(list("ab"))),
                (("a"), set(list("a"))),
                ((""), set()),
                (("    "), set()),
                (("a, b, c"), set(list("abc"))),
                ((" a , b , c "), set(list("abc"))),
                ((" a,b, c"), set(list("abc"))),
            ],
        ] * 12


@attr(test="unit")
class TestAutoSqlDeclaration(TestAbstractAutoSqlElement):
    @classmethod
    def setUpClass(cls):
        # define all of these in subclasses
        cls.test_class = AutoSqlDeclaration
        cls.match_key = cls.test_class.__name__

        # instances of autoSql text that match cls.test_class.match_pattern
        cls.matching_text = match_texts[cls.match_key]

        # instances of autoSql text that do not match cls.test_class.match_pattern
        cls.non_matching_text = list(
            itertools.chain(
                *map(lambda x: x[1], filter(lambda x: x[0] != cls.match_key, match_texts.items()))
            )
        )

        cls.test_record = OrderedDict()

        # list of lists of tuples of unparsed text and expected result objects
        # for each autoSql declaration in cls.matching_text
        cls.expected_values = [
            [
                (
                    (
                        "3    1,2,3    my string with spaces    5    1.1,1.2,1.3,1.4,1.5    a,b,c".
                        replace("    ", "\t")
                    ),
                    OrderedDict(
                        [
                            ("number", 3), ("points", (1, 2, 3)),
                            ("my_string", "my string with spaces"), ("a_field_size", 5),
                            ("float_array", (1.1, 1.2, 1.3, 1.4, 1.5)), ("alpha", set(list("abc")))
                        ]
                    )
                ),
                (
                    (
                        "200    5, 212341,12341323    my string with spaces and @&$@(#!@&!    2    3.0,-4586.2    a".
                        replace("    ", "\t")
                    ),
                    OrderedDict(
                        [
                            ("number", 200), ("points", (5, 212341, 12341323)),
                            ("my_string", "my string with spaces and @&$@(#!@&!"),
                            ("a_field_size", 2), ("float_array",
                                                  (3.0, -4586.2)), ("alpha", set(list("a")))
                        ]
                    )
                ),
                (
                    (
                        "200    5, 212341,12341323    my string with spaces and @&$@(#!@&!    2    3.0,-4586.2    ".
                        replace("    ", "\t")
                    ),
                    OrderedDict(
                        [
                            ("number", 200), ("points", (5, 212341, 12341323)),
                            ("my_string", "my string with spaces and @&$@(#!@&!"),
                            ("a_field_size", 2), ("float_array", (3.0, -4586.2)), ("alpha", set())
                        ]
                    )
                ),
            ]
        ] * 3

    def test_warning_on_bad_type(self):
        bad_table = \
"""table easy_table "A table with a comment on the next line" ( 
uint number auto; "a number with a token"
uint [3] points ; "r,g,b values"
lstring  my_string ; "a long string"
uint a_field_size ; "the size for the next field"
float [a_field_size] float_array ; "an array of floats"
set(a,b,c) alpha ; "the first three letters of the alphabet"
stringz  badtype ; "non-existent type!")
"""
        # reset warning registry temporarily to prevent weird interactions with nose
        global pl_once_registry
        tmp = pl_once_registry
        tmp2 = warnings.filters
        warnings.filters = []
        pl_once_registry = {}
        with warnings.catch_warnings(record=True) as warns:
            AutoSqlDeclaration(bad_table)

        pl_once_registry = tmp
        warnings.filters = tmp2
        self.assertEqual(
            1, len(warns),
            "AutoSqlDeclaration() failed to warn on bad type name (warned %s times)." % len(warns)
        )

    def test_warning_on_double_declaration(self):
        bad_table = \
"""table easy_table "A table with a comment on the next line" ( 
uint number auto; "a number with a token"
uint [3] points ; "r,g,b values"
lstring  my_string ; "a long string"
uint a_field_size ; "the size for the next field"
float [a_field_size] float_array ; "an array of floats"
set(a,b,c) alpha ; "the first three letters of the alphabet"
string  number ; "a duplicated field!")
"""
        # reset warning registry temporarily to prevent weird interactions with nose
        global pl_once_registry
        tmp = pl_once_registry
        tmp2 = warnings.filters
        warnings.filters = []
        pl_once_registry = {}
        with warnings.catch_warnings(record=True) as warns:
            AutoSqlDeclaration(bad_table)

        pl_once_registry = tmp
        warnings.filters = tmp2
        self.assertEqual(
            1, len(warns),
            "AutoSqlDeclaration() failed to warn on duplicated field name (warned %s times)." %
            len(warns)
        )



#!/usr/bin/env python
"""Tools for manipulating lists

Methods
-------
:py:func:`flatten_nested_lists_to_generator`
    Flatten nested lists, from left-to-right and top-to-bottom, into a generator

:py:func:`flatten_nested_lists_to_list`
    Flatten nested lists, from left-to-right and top-to-bottom, into a list

:py:func:`parse_list`
    Parse string representation of list into a list of parsed data (e.g.
    a list of numbers, a list of strings, et c)
"""

from plastid.util.services.misc import guess_formatter


def parse_list(inp):
    """Restore non-nested lists of Python primitives from string representation of list
    Additionally parses `numpy.nan`, `numpy.inf`, and `-numpy.inf` to correct values
    
    Parameters
    ----------
    inp : str
        String representation of list data

    Returns
    -------
    list

    Raises
    ------
    AssertionError
        if String does not represent a list

    Notes
    -----
    Complex types like nested lists or tuples will not be processed,
    and may not be parsed correctly.
    """
    assert inp[0] == '[' and inp[-1] == ']'
    return [guess_formatter(X.strip().strip("'")) for X in inp[1:-1].split(",")]


# Adapted from http://stackoverflow.com/questions/2158395/flatten-an-irregular-list-of-lists-in-python
def flatten_nested_lists_to_generator(l):
    """Flatten a tree of nested lists into a generator, from left-to-right and top-to-bottom.
        
    
    Parameters
    ----------
    l : list
        A tree as a list of lists
    
    Yields
    ------
    object
        Next item in flattened list
    """
    assert isinstance(l, list)
    for el in l:
        if isinstance(el, list):
            for sub in flatten_nested_lists_to_generator(el):
                yield sub
        else:
            yield el


def flatten_nested_lists_to_list(inp):
    """Flatten a tree of lists into a single list of items, from left-to-right and top-to-bottom.
    
    Parameters
    ----------
    inp : list
        A tree as a list of lists
    
    Returns
    -------
    list
        flattened list

    Notes
    -----
    Tuples and other sequences will not be flattened. They will remain in
    their native types
    """
    return list(flatten_nested_lists_to_generator(inp))



from collections import OrderedDict # cimport?

from libc.stddef cimport size_t
from plastid.genomics.c_common cimport Strand, forward_strand, reverse_strand, unstranded
from plastid.genomics.roitools cimport GenomicSegment

cdef:
    str WARN_CHROM_NOT_FOUND

#===============================================================================
# INDEX: Common info for all bigwig files
#===============================================================================

DEF bigBedSig = 0x8789F2EB
DEF bigWigSig = 0x888FFC26

cdef extern from "<common.h>" nogil:
    ctypedef unsigned char   Bits
    ctypedef unsigned char   bits8
    ctypedef unsigned short  bits16
    ctypedef unsigned bits32
    ctypedef unsigned long long bits64

    void freeMem(void *pt)
#
#     cdef struct fileOffsetSize:
#         fileOffsetSize * next
#         bits64           offset
#         bits64           size



cdef extern from "<bits.h>":

    # find index of next set bit
    int bitFindSet(Bits *b, int startIx, int bitCount)

    # find index of next clear bit
    int bitFindClear(Bits *b, int startIX, int bitCount)

    # read next bit
    bint bitReadOne(Bits *b, int bitIx)


cdef extern from "<localmem.h>":
    cdef struct lm:
        lmBlock *blocks
        size_t blockSize
        size_t allignMask
        size_t allignAdd

    cdef struct lmBlock:
        lmBlock *next
        char *free
        char *end
        char *extra

    lm * lmInit(int blockSize)
    size_t lmAvailable(lm *lm)
    size_t lmSize(lm *lm)
    void * lmAlloc(lm *lm, size_t size)
    void * lmAllocMoreMem(lm *lm, void *pt, size_t oldSize, size_t newSize)
    void lmCleanup(lm **pLm)


cdef extern from "<bbiFile.h>":
    cdef struct bbiZoomLevel:
        bbiZoomLevel *next
        bits32 reductionLevel
        bits64 dataOffset
        bits64 indexOffset

    cdef struct bbiFile:
        bbiFile *next
        char *fileName
        #struct udcFile *udc
        bits32 typeSig
        bint   isSwapped
        #struct bptFile *chromBpt
        bits16 version
        bits16 zoomLevels
        #bits64 chromTreeOffset
#         bits64 unzoomedDataOffset
#         bits64 unzoomedIndexOffset
        bits16 fieldCount
        bits16 definedFieldCount
        bits64 asOffset
        bits64 totalSummaryOffset
        bits32 uncompressBufSize
        #bits64 extensionOffset
        #struct cirTreeFile unzoomedCir
        bbiZoomLevel *levelList
        #bits16 extensionSize
        bits16 extraIndexCount
        #bits64 extraIndexListOffset

    cdef struct bbiChromIdSize:
        bits32 chromId
        bits32 chromSize

    cdef struct bbiChromInfo:
        bbiChromInfo * next
        char *name
        bits32 id
        bits32 size

    cdef struct bbiSummary:
        bbiSummary * next_
        bits32 chromId
        bits32 start, end
        bits32 validCount
        float minVal
        float maxVal
        float sumData
        float sumSquares
        bits64 fileOffset

    cdef struct bbiSummaryElement:
        bits64 validCount
        double minVal
        double maxVal
        double sumData
        double sumSquares

    cdef enum bbiSummaryType:
        bbiSumMean              = 0
        bbiSumMax               = 1
        bbiSumMin               = 2
        bbiSumCoverage          = 3
        bbiSumStandardDeviation = 4

    # bbiFile *bbiFileOpen(char *fileName, bits32 sig, char *typeName)

    void bbiFileClose(bbiFile **pBwf)

    bbiChromInfo *bbiChromList(bbiFile *bbi)
    void bbiChromInfoFreeList(bbiChromInfo **pList)

    char *bbiCachedChromLookup(bbiFile *bbi,
                               int chromId,
                               int lastChromId,
                               char *chromBuf,
                               int chromBufSize)

    bbiSummaryElement bbiTotalSummary(bbiFile *bbi)


# cdef extern from "<cirTree.h>":
#     cdef struct cirTreeFile:
#     # R tree index file handle. */
#         struct cirTreeFile *next;    # Next in list of index files if any. */
#         char *fileName;              # Name of file - for error reporting. */
#         #struct udcFile *udc;         # Open file pointer. */
#         boolean isSwapped;           # If TRUE need to byte swap everything. */
#         bits64 rootOffset;           # Offset of root block. */
#         bits32 blockSize;            # Size of block. */
#         bits64 itemCount;            # Number of items indexed. */
#         bits32 startChromIx;         # First chromosome in file. */
#         bits32 startBase;            # Starting base position. */
#         bits32 endChromIx;           # Ending chromosome in file. */
#         bits32 endBase;              # Ending base position. */
#         bits64 fileSize;             # Total size of index file. */
#         bits32 itemsPerSlot;         # Max number of items to put in each index slot at lowest level. */
#
#         struct fileOffsetSize *cirTreeEnumerateBlocks(struct cirTreeFile *crf)




# Close a BigBed/BigWig file
cdef inline void close_file(bbiFile *myFile):
    bbiFileClose(&myFile)

# Allocate or destroy and reallocate a local memory block of size <= maxmem
cdef lm * get_lm(lm * my_lm=*, int maxmem=*) except NULL


#===============================================================================
# INDEX: Python class
#===============================================================================


cdef class _BBI_Reader:

    cdef:
        bbiFile * _bbifile
        str       filename
        dict      _chromids
        dict      _chromlengths
        dict      _summary
        lm *      _lm
        long      _maxmem

#         dict _zoomlevels
#         dict offsets
#         bits64 _unzoomed_data_offset, _unzoomed_index_offset

    cdef dict _define_chroms(self)
    cdef dict c_chroms(self)
    cdef lm* _get_lm(self)
#     cdef dict fetch_summary(self)



#!/usr/bin/env python


class UniqueFIFO(object):
    """FIFO of unique objects. In other words, if an element already present
    in the FIFO is appended to the FIFO, it is moved to the right end and 
    no element is removed from the FIFO. Elements are only removed when a 
    element not present in the FIFO is appended to the right end, and when
    the number of elements in the FIFO exceeds `self.max_size`.

        
    Parameters
    ----------
    idx : int
        Index of item to fetch
    
    Returns
    -------
    object
        Nth item in the |UniqueFIFO|
        
    
    Attributes
    ----------
    max_size : int
        Maximum size of |UniqueFIFO|
    """

    def __init__(self, size):
        assert size > 0
        self.max_size = size
        self._elements = []

    def __getitem__(self, idx):
        """Fetch Nth item in the |UniqueFIFO|
        
        Parameters
        ----------
        idx : int
            Index of item to fetch
        
        Returns
        -------
        object
            Nth item in the |UniqueFIFO|
        """
        return self._elements[idx]

    def __contains__(self, el):
        """Determine whether an element is in the |UniqueFIFO|
        
        Parameters
        ----------
        el : hashable object
            Query object
        
        Returns
        -------
        bool
            `True` if `el` is in the |UniqueFIFO|, `False` otherwise
        """
        return el in self._elements

    def __iter__(self):
        """Iterate over elements in the |UniqueFIFO|"""
        return iter(self._elements)

    def __len__(self):
        """Return number of objects presently in the |UniqueFIFO|
        
        Returns
        -------
        int
            Number of elements presently in the |UniqueFIFO|, must be less than
            or equal to `self.max_size`
        """
        return len(self._elements)

    def append(self, el):
        """Append an item to the |UniqueFIFO|. If the item is already present
        in the |UniqueFIFO|, it is moved to the right end of the FIFO, and the
        length of the |UniqueFIFO| is unchanged. Otherwise, the element is
        appended to the right end. If the length of the |UniqueFIFO| then 
        exceeds `self.max_size`, the fist element is popped.
        
        Parameters
        ----------
        el : hashable
            Any hashable object
        """
        if el in self:
            # move to front if present
            n = self._elements.index(el)
            self._elements = self._elements[:n] + self._elements[n + 1:] + [el]
        else:
            # otherwise append, removing first element
            # if we are too long
            if len(self._elements) >= self.max_size:
                self._elements = self._elements[1:]
                self._elements.append(el)
            else:
                self._elements.append(el)

    def __str__(self):
        return str(self._elements)

    def __repr__(self):
        return "<UniqueFifo %s>" % repr(self._elements)



cdef str strand_to_str(Strand strand):
    """Convert enum Strand to str representation"""
    if strand == forward_strand:
        return "+"
    elif strand == reverse_strand:
        return "-"
    elif strand == unstranded:
        return "."
    elif strand == undef_strand:
        return "strand undefined"
    else:
        raise ValueError(
            "strand_to_str: Strand must be forward (%s), reverse (%s), or unstranded(%s). Got '%s'."
            % (forward_strand, reverse_strand, unstranded, strand)
        )

cdef Strand str_to_strand(str val) except error_strand:
    """Convert str representation of strand to enum"""
    if val == "+":
        return forward_strand
    elif val == "-":
        return reverse_strand
    elif val == ".":
        return unstranded
    elif val == "\x00":
        return undef_strand
    else:
        raise ValueError("Strand must be '+', '-', '.', or '\\x00' (undefined). Got '%s'." % val)

cdef class _GeneratorWrapper(object):
    """Wrapper class to prevent `repr()` of Cython generators from erroring in
    interactive sessions.  All attributes and behaviors are delegated to the
    wrapped generator, except the attribute `generator`.

    Attributes
    ----------
    generator : generator
        Wrapped generator object
    """

    def __cinit__(self, generator, desc):
        """Wrapper class to prevent `repr()` of Cython generators from erroring
        in interactive sessions

        Parameters
        ----------
        generator : Generator object to wrap

        desc : str
            Short description of generator contents, used in generating
            the  `repr` text
        """
        self.generator = generator
        self.desc = "<wrapped generator object '%s'>" % desc

    def __str__(self):
        return self.desc

    def __repr__(self):
        return self.desc

    def __getattr__(self, attr):
        if attr == "generator":
            return self.generator
        else:
            return getattr(self.generator, attr)

    def __iter__(self):
        return iter(self.generator)

    def __next__(self):
        return next(self.generator)

    def next(self):
        return next(self.generator)



#!/usr/bin/env python
"""Tests capabilities of |UniqueFIFO|"""
import unittest
from nose.plugins.attrib import attr

from plastid.util.unique_fifo import UniqueFIFO
from collections import Counter


@attr(test="unit")
class TestUniqueFIFO(unittest.TestCase):
    def test_getitem(self):
        my_fifo = UniqueFIFO(5)
        for i in range(5):
            my_fifo.append(i)
            self.assertEqual(i, my_fifo[-1])

        for i in range(5):
            self.assertEqual(i, my_fifo[i])

        for i in range(10):
            my_fifo.append(i)
            self.assertEqual(i, my_fifo[-1])

    def test_contains(self):
        my_fifo = UniqueFIFO(5)
        self.assertFalse(0 in my_fifo)
        my_fifo.append(0)
        self.assertTrue(0 in my_fifo)
        my_fifo.append(1)
        self.assertTrue(0 in my_fifo)
        my_fifo.append(2)
        self.assertTrue(0 in my_fifo)
        my_fifo.append(3)
        self.assertTrue(0 in my_fifo)
        my_fifo.append(4)
        self.assertTrue(0 in my_fifo)
        my_fifo.append(5)
        self.assertFalse(0 in my_fifo)

    def test_iter(self):
        fifo_items1 = list(range(5, 10))
        fifo_items2 = list(range(10, 15))
        all_fifo_items = fifo_items1 + fifo_items2
        my_fifo = UniqueFIFO(5)
        # add up to length, and make sure at each addition
        # the right element is in each spot
        for item in fifo_items1:
            my_fifo.append(item)
            for c, item in enumerate(my_fifo):
                self.assertEqual(item, fifo_items1[c])

        # add at sizes >= max_length, and make sure at each addition
        # the right element is in each spot
        for n, item in enumerate(fifo_items2):
            my_fifo.append(item)
            for c, item in enumerate(my_fifo):
                self.assertEqual(item, all_fifo_items[n + c + 1])

    def test_len(self):
        my_fifo = UniqueFIFO(5)
        for i in range(5):
            self.assertEqual(len(my_fifo), i)
            my_fifo.append(i)
        for i in range(5):
            self.assertEqual(len(my_fifo), 5)

    def test_append_no_exceed_maxlen(self):
        for max_size in range(1, 10):
            my_fifo = UniqueFIFO(max_size)
            for i in range(100):
                my_fifo.append(i)
                self.assertTrue(len(my_fifo) <= max_size)
                self.assertTrue(len(list(my_fifo)) <= max_size)

    def test_append_up_to_maxlen(self):
        my_fifo = UniqueFIFO(5)
        for i in range(5):
            my_fifo.append(i)
            self.assertEqual(len(my_fifo), i + 1)

    def test_append_reorder(self):
        my_fifo = UniqueFIFO(5)
        my_fifo.append(0)
        my_fifo.append(1)
        my_fifo.append(2)
        my_fifo.append(3)
        my_fifo.append(4)
        self.assertListEqual(list(my_fifo), [0, 1, 2, 3, 4])

        my_fifo.append(3)
        self.assertListEqual(list(my_fifo), [0, 1, 2, 4, 3])

        my_fifo.append(0)
        self.assertListEqual(list(my_fifo), [1, 2, 4, 3, 0])

        my_fifo.append(4)
        self.assertListEqual(list(my_fifo), [1, 2, 3, 0, 4])

        my_fifo.append(2)
        self.assertListEqual(list(my_fifo), [1, 3, 0, 4, 2])

        my_fifo.append(2)
        self.assertListEqual(list(my_fifo), [1, 3, 0, 4, 2])

        my_fifo.append(2)
        self.assertListEqual(list(my_fifo), [1, 3, 0, 4, 2])

        my_fifo.append(2)
        self.assertListEqual(list(my_fifo), [1, 3, 0, 4, 2])

    def test_append_no_duplicates(self):
        my_fifo = UniqueFIFO(5)
        for i in range(5):
            my_fifo.append(i)
        self.assertListEqual(list(set(my_fifo)), list(my_fifo))
        for i in range(20):
            my_fifo.append(i)
            for v in Counter(my_fifo).values():
                self.assertEqual(v, 1)
